# -*- coding: utf-8 -*-
"""
This script produces Figure G.3 in the appendix of "Asymptotically Valid
Bootstrap Inference for Proxy SVARs" by Carsten Jentsch and Kurt G. Lunsford.

This script saves
    FigureApp3.pdf,
shown as Figure G.3 in the appendix.

This script calls the following Excel file:
    dgp1_results68.xlsx
This Excel file is produced by appendix_script.py, which must be run before
this script can be used.

This script uses numpy, pandas, and matplotlib packages.
"""

#%% Import Packages and Functions

# Pandas, NumPy, and Matplotlib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


#%% Parameters of the Script

# These parameters must match what is in the Excel files
K = 2           # Number of variables in the VAR
n_imp = 21      # Horizon for IRFs and FEVDs, including the impact (0 horizon)


#%% Load the Data

df_parms = pd.read_excel('dgp1_results68.xlsx', 'parameters', index_col=0)
df_irfnorm = pd.read_excel('dgp1_results68.xlsx', 'length', index_col=0)


#%% Parameters to Print

psi1 = df_parms.loc[0, 'psi']
n_sample = df_parms.loc[0, 'n_sample']
ell = df_parms.loc[0, 'ell']


# %% Collect the Results

plot_data = np.zeros((K, n_imp, 6))
for jj in range(6):
    plot_data[:, :, jj] = df_irfnorm.iloc[:, jj*K:(jj+1)*K].T


#%% Plot the Results

# Variables for the x-axes
x_vars = np.arange(n_imp)
x_tick = np.linspace(0, n_imp-1, int((n_imp-1)/4))

# Markers to be used in the plots
marker_choices = ['x', 'o', 'v', 's', '*', '^']

# Figure and axes instances
fig, ax= plt.subplots(1, 2, figsize=(10, 5))

# First variable with dgp1
for ii in range(6):
    ax[0].plot(x_vars, plot_data[0, :, ii], linewidth=2, linestyle="-",
               marker=marker_choices[ii], markerfacecolor='None', 
               markeredgewidth=2)
ax[0].set_xlim(0, n_imp-1)
ax[0].set_ylim(0, 1)
ax[0].set_xticks(x_tick)
ax[0].set_yticks(np.linspace(0, 1, 6))
ax[0].set_xlabel('Horizon', fontsize=12)
ax[0].set_ylabel('Average Length', fontsize=12)
t = ('$y_{1,t}, T=' + str(n_sample) + ', \ell=' + str(ell) 
     + ', \psi=' + str(psi1) + '$')
ax[0].set_title(t, fontsize=12)

# Second variable with dgp1
for ii in range(6):
    ax[1].plot(x_vars, plot_data[1, :, ii], linewidth=2, linestyle="-",
               marker=marker_choices[ii], markerfacecolor='None', 
               markeredgewidth=2)
ax[1].set_xlim(0, n_imp-1)
ax[1].set_ylim(0, 1)
ax[1].set_xticks(x_tick)
ax[1].set_yticks(np.linspace(0, 1, 6))
ax[1].set_xlabel('Horizon', fontsize=12)
ax[1].set_ylabel('Average Length', fontsize=12)
t = ('$y_{2,t}, T=' + str(n_sample) + ', \ell=' + str(ell) 
     + ', \psi=' + str(psi1) + '$')
ax[1].set_title(t, fontsize=12)

ax[1].legend(['MBB', 'grid MBB AR', 'MSW', 'WB-Rademacher', 'MBB Halls', 'grid Sim AR'],
             ncol=6, loc='center', bbox_to_anchor=(-0.1, -0.2), fontsize=12)

plt.savefig('FigureApp3.pdf', bbox_inches='tight')

