# -*- coding: utf-8 -*-
"""
This script produces results shown in Appendix G.6 of "Asymptotically Valid
Bootstrap Inference for Proxy SVARs" by Carsten Jentsch and Kurt G. Lunsford.

It produces results for one of three data generating processes.  Each data
generating process is differentiated by proxy strength.  The results are
automatically saved in two of the following Excel files:
    dgp1_results68.xlsx
    dgp1_results95.xlsx
    dgp2_results68.xlsx
    dgp2_results95.xlsx
    dgp3_results68.xlsx
    dgp3_results95.xlsx
These files are differentiated by data generating process and nominal level of
confidence intervals and sets (either 68 or 95).  When all six sets of results
are produced (two for each data generating process) and saved, they are called
by other scripts to produce figures in Appendix G.6.

This script uses functions from the following modules:
    proxy_svar_module.py
    utilities_module.py

This script also uses numpy, pandas, os, sys, and time packages.
"""

#%% Import Packages and Functions

# Pandas and NumPy
import pandas as pd
import numpy as np

# Get replication files directory
import os
path_dir = os.path.dirname(os.getcwd())

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Functions for appendix simulations
from modules.proxy_svar_module import make_dynamics, make_proxy_svar_dynamics
from modules.proxy_svar_module import make_proxy_svar_mbb, make_proxy_svar_wbr
from modules.proxy_svar_module import make_msw_cs, make_halls, make_ar_sim
from modules.utilities_module import make_data_sv, check_include

# Time the program
import time
start = time.time()


#%% Choose the Data Generating Process

# Notes:
#   Set equal to 1 for psi = 1.00
#   Set equal to 2 for psi = 0.32
#   Set equal to 3 for psi = 0.17
chooseDGP = 1


#%% Parameters of the Script

# Parameters of the SVAR
A0 = np.zeros((2, 1))
A1 = np.array([(0.44, 0.66), (-0.11, 1.32)])
A2 = np.array([(-0.18, 0), (-0.18, -0.09)])
A = np.hstack((A0, A1, A2))
H = np.array([(np.sqrt(2)/2, np.sqrt(2)/2),
              ((np.sqrt(2)-np.sqrt(6))/4, (np.sqrt(2)+np.sqrt(6))/4)])

# Parameters of stochastic volatility
rho = 0.95
sigma_e = 0.09
log_sig_bar = -sigma_e*sigma_e/(1 - rho*rho)

# Covariance between structural shock of interest and proxy variable
if chooseDGP == 1:
    psi = 1.00          # Value for strong proxy DGP
elif chooseDGP == 2:
    psi = 0.32          # Value for medimum proxy DGP
elif chooseDGP == 3:
    psi = 0.17          # Value for weak proxy DGP
else:
    print('Error!')
    print('"chooseDGP" must be set to either 1, 2 or 3.')
    exit()

# Population mean of the proxy variable
m_bar = 0

# Fraction of proxy variables that are censored
exp_d = 0

# IRF and FEVD parameters
s = -1          # scale for the normalized IRF
n_imp = 21      # horizon for IRFs and FEVDs, including the impact (0 horizon)

# Simulation parameters
n_sample = 200          # sample sizes to be used
burn = 1000             # number of observations to be burned
blocksize = 4           # block size for each sample size to be used
nBoot = 2000            # number of bootstrap replications
nSim = 1000              # number of simulations

# Seed the random number generator
np.random.seed(0)


#%% Store the Parameters in a DataFrame

cols_parm = ['n_sample', 'ell', 'psi']
df_parm = pd.DataFrame(index=[0], columns=cols_parm)

df_parm.loc[:, 'n_sample'] = n_sample
df_parm.loc[:, 'ell'] = blocksize
df_parm.loc[:, 'psi'] = psi


#%% Compute Additional Parameters and the Population IRFs and FEVDs

# Compute dimensions of the data generating process
K = A[:, 0].size                # Number of variables in the VAR
p = int((A[0, :].size - 1)/K)   # Number of lags of the VAR

# Compute some population parameters
H1 = H[:, 0:1]                  # First column of H
SigmaU = H @ H.T                # Covariance matrix of VAR innovations

# Compute IRFs and FEVDs
IRF_true, IRFnorm_true, FEVD_true, SVMA_true \
    = make_dynamics(A, SigmaU, H1, H1, p, s, n_imp)


#%% Set the Grid for the Anderson Rubin Grid Bootstrap

# Construct the grid include the values in the true IRF
ARgrid = np.empty(K*n_imp)
ARgrid[:] = IRFnorm_true.flatten()

# Add extreme values to the grid check if the confidence set is "ubounded"
ARgrid = np.append(ARgrid, np.array([-1000, 1000]))

# Add additional points to get a sense of the length of the confidence set
if chooseDGP == 1:
    temp = np.abs(s)*np.arange(-3, 7, 0.005)
    ARgrid = np.append(ARgrid, temp)


#%% Storage Arrays for Simulation Results

# Storage arrays for moving block bootstrap percentile intervals
irf_include68 = np.zeros((K, n_imp, nSim))
irf_norm_include68 = np.zeros((K, n_imp, nSim))
fevd_include68 = np.zeros((K, n_imp, nSim))
irf_include95 = np.zeros((K, n_imp, nSim))
irf_norm_include95 = np.zeros((K, n_imp, nSim))
fevd_include95 = np.zeros((K, n_imp, nSim))

# Storage arrays for Montiel Olea, Stock and Watson analytic confidence sets
msw_include68 = np.zeros((K, n_imp, nSim))
msw_include95 = np.zeros((K, n_imp, nSim))

# Storage arrays for Anderson Rubin type grid confidence set
ar_mbb_include68 = np.zeros((K, n_imp, nSim))
ar_extreme_include68 = np.zeros((K, n_imp, nSim))
ar_mbb_include95 = np.zeros((K, n_imp, nSim))
ar_extreme_include95 = np.zeros((K, n_imp, nSim))

# Storage arrays for Anderson Rubin type grid confidence set with simulation
ar_sim_include68 = np.zeros((K, n_imp, nSim))
ar_extreme_sim_include68 = np.zeros((K, n_imp, nSim))
ar_sim_include95 = np.zeros((K, n_imp, nSim))
ar_extreme_sim_include95 = np.zeros((K, n_imp, nSim))

# Storage arrays for moving block bootstrap Hall's percentile intervals
irf_include_halls68 = np.zeros((K, n_imp, nSim))
irf_norm_include_halls68 = np.zeros((K, n_imp, nSim))
fevd_include_halls68 = np.zeros((K, n_imp, nSim))
irf_include_halls95 = np.zeros((K, n_imp, nSim))
irf_norm_include_halls95 = np.zeros((K, n_imp, nSim))
fevd_include_halls95 = np.zeros((K, n_imp, nSim))

# Storage arrays for Rademacher wild bootstrap percentile intervals
irf_include_wbr68 = np.zeros((K, n_imp, nSim))
irf_norm_include_wbr68 = np.zeros((K, n_imp, nSim))
fevd_include_wbr68 = np.zeros((K, n_imp, nSim))
irf_include_wbr95 = np.zeros((K, n_imp, nSim))
irf_norm_include_wbr95 = np.zeros((K, n_imp, nSim))
fevd_include_wbr95 = np.zeros((K, n_imp, nSim))

# Storage arrays for F statistics
f_store = np.zeros(nSim)
f_iv_store = np.zeros(nSim)
w_store = np.zeros(nSim)

# Storage arrays to measure the length confidence intervals/sets
if chooseDGP == 1:
    len_mbb68 = np.zeros((K, n_imp, nSim))
    len_mbb_ar68 = np.zeros((K, n_imp, nSim))
    len_msw68 = np.zeros((K, n_imp, nSim))
    len_wbr68 = np.zeros((K, n_imp, nSim))
    len_mbb_halls68 = np.zeros((K, n_imp, nSim))
    len_sim_ar68 = np.zeros((K, n_imp, nSim))
    
    len_mbb95 = np.zeros((K, n_imp, nSim))
    len_mbb_ar95 = np.zeros((K, n_imp, nSim))
    len_msw95 = np.zeros((K, n_imp, nSim))
    len_wbr95 = np.zeros((K, n_imp, nSim))
    len_mbb_halls95 = np.zeros((K, n_imp, nSim))
    len_sim_ar95 = np.zeros((K, n_imp, nSim))


#%% Simulation

for sim in range(nSim):
    # Construct the observable time series
    y, shocks, vv = make_data_sv((burn+np.max(n_sample)), K, p, A, H,
                                 rho, log_sig_bar, sigma_e)
    
    # Drop the burn period less the initial p observations of VAR data
    y = y[burn-p:, :]
    
    # Drop the burn period of the proxy variable components
    shocks = shocks[burn-p:, :]
    vv = vv[burn-p:, :]
    
    # Compute the censoring
    d = (np.random.uniform(size=((n_sample+p), 1)) > exp_d)
    
    # Set the data sample
    var_data = np.zeros((n_sample+p, K))
    var_data = y[0:n_sample+p, :]
    
    # Construct the proxy variable
    proxies = np.zeros((n_sample+p, 1))
    proxies[:, :] = m_bar + psi*shocks[0:n_sample+p, 0:1] + vv[0:n_sample+p, :]
    proxies[:, :] = np.multiply(proxies, d[0:n_sample+p, :])
    
    # Estimate the proxy SVAR and compute the dynamics
    SVAR_results = make_proxy_svar_dynamics(var_data, proxies, p, s, n_imp)
    
    # Save some estimates for additional calculations
    a_hat = np.zeros((K, 1+K*p))
    a_hat[:, :] = SVAR_results['a_hat']
    
    u_hat = np.zeros((n_sample, K))
    u_hat[:, :] = SVAR_results['u_hat']
    
    svma_hat = np.zeros((K, n_imp))
    svma_hat = SVAR_results['svma_hat']
    
    # Set the initial conditions for the bootstrap
    y0 = np.zeros(1+K*p)
    y0[0] = 1
    for ii in range(p):
        y0[1+p*ii:1+p*(ii+1)] = var_data[p-1-ii, :]
    
    # Truncate the proxy variables to align with the VAR residuals
    proxies_trunc = np.zeros((n_sample, 1))
    proxies_trunc[:, 0] = proxies[p:n_sample+p, 0]
    
    # Compute moving block bootstrap CIs
    MBB_results = make_proxy_svar_mbb(a_hat, u_hat, proxies_trunc, y0,
                                      p, s, n_imp, nBoot, blocksize,
                                      toggleAR=1, ar_grid=ARgrid,
                                      svma_vals=svma_hat)
    
    # Compute Monteil Olea, Stock and Watson confidence sets
    MSW_results = make_msw_cs(a_hat, var_data, u_hat, proxies_trunc,
                              p, s, n_imp)
    
    # Compute rademacher wild bootstrap CIs
    WBR_results = make_proxy_svar_wbr(a_hat, u_hat, proxies_trunc, y0,
                                      p, s, n_imp, nBoot)
    
    # Change moving block bootstrap percentile intervals to Hall's intervals
    ci68halls_irf \
        = make_halls(MBB_results["ci68_irf"], SVAR_results["irf_hat"])
    ci68halls_irf_norm \
        = make_halls(MBB_results["ci68_irf_norm"], SVAR_results["irf_norm_hat"])
    ci68halls_fevd \
        = make_halls(MBB_results["ci68_fevd"], SVAR_results["fevd_hat"])
    
    ci95halls_irf \
        = make_halls(MBB_results["ci95_irf"], SVAR_results["irf_hat"])
    ci95halls_irf_norm \
        = make_halls(MBB_results["ci95_irf_norm"], SVAR_results["irf_norm_hat"])
    ci95halls_fevd \
        = make_halls(MBB_results["ci95_fevd"], SVAR_results["fevd_hat"])
    
    # Compute AR-type sets with asymptotic simulation
    ARSim_results = make_ar_sim(var_data, proxies, p, s, n_imp, nBoot, ARgrid)
    
    # Check if moving block bootstrap CIs include true values
    irf_include68[:, :, sim] \
        = check_include(K, n_imp, MBB_results["ci68_irf"], IRF_true)
    irf_norm_include68[:, :, sim] \
        = check_include(K, n_imp, MBB_results["ci68_irf_norm"], IRFnorm_true)
    fevd_include68[:, :, sim] \
        = check_include(K, n_imp, MBB_results["ci68_fevd"], FEVD_true)
    
    irf_include95[:, :, sim] \
        = check_include(K, n_imp, MBB_results["ci95_irf"], IRF_true)
    irf_norm_include95[:, :, sim] \
        = check_include(K, n_imp, MBB_results["ci95_irf_norm"], IRFnorm_true)
    fevd_include95[:, :, sim] \
        = check_include(K, n_imp, MBB_results["ci95_fevd"], FEVD_true)
    
    # Check if normalized IRFs are included in the AR grid confidence sets
    grid = MBB_results["ar_grid"]
    ar_mbb_set68 = MBB_results["ar_index68"]
    ar_mbb_set95 = MBB_results["ar_index95"]
    counter = 0
    for kk in range(K):
        for ii in range(n_imp):
            ar_mbb_include68[kk, ii, sim] = ar_mbb_set68[counter, ii, kk]
            ar_mbb_include95[kk, ii, sim] = ar_mbb_set95[counter, ii, kk]
            counter = counter + 1
    
    # Check if extreme values are included in AR grid confidence sets
    for kk in range(K):
        for ii in range(n_imp):
            ar_extreme_include68[kk, ii, sim] \
                = np.all([ar_mbb_set68[2*n_imp, ii, kk],
                          ar_mbb_set68[2*n_imp+1, ii, kk]])
            ar_extreme_include95[kk, ii, sim] \
                = np.all([ar_mbb_set95[2*n_imp, ii, kk],
                          ar_mbb_set95[2*n_imp+1, ii, kk]])
    
    # Check if normalized IRFs are included in the MSW sets
    cs68_msw = MSW_results["cs68_irf_norm"]
    if cs68_msw[:, 0, 0].size == 2:
        msw_include68[:, :, sim] \
            = check_include(K, n_imp, cs68_msw, IRFnorm_true)
    else:
        for kk in range(K):
            for ii in range(n_imp):
                temp = np.empty((4, 1))
                temp[0, 0] = cs68_msw[0, ii, kk] <= IRFnorm_true[kk, ii]
                temp[1, 0] = IRFnorm_true[kk, ii] <= cs68_msw[1, ii, kk]
                temp[2, 0] = cs68_msw[2, ii, kk] <= IRFnorm_true[kk, ii]
                temp[3, 0] = IRFnorm_true[kk, ii] <= cs68_msw[3, ii, kk]
                msw_include68[kk, ii, sim] \
                    = np.all(temp[0:2, 0]) + np.all(temp[2:4, 0])
    
    cs95_msw = MSW_results["cs95_irf_norm"]
    if cs95_msw[:, 0, 0].size == 2:
        msw_include95[:, :, sim] \
            = check_include(K, n_imp, cs95_msw, IRFnorm_true)
    else:
        for kk in range(K):
            for ii in range(n_imp):
                temp = np.empty((4, 1))
                temp[0, 0] = cs95_msw[0, ii, kk] <= IRFnorm_true[kk, ii]
                temp[1, 0] = IRFnorm_true[kk, ii] <= cs95_msw[1, ii, kk]
                temp[2, 0] = cs95_msw[2, ii, kk] <= IRFnorm_true[kk, ii]
                temp[3, 0] = IRFnorm_true[kk, ii] <= cs95_msw[3, ii, kk]
                msw_include95[kk, ii, sim] \
                    = np.all(temp[0:2, 0]) + np.all(temp[2:4, 0])
    
    # Check if Rademacher wild bootstrap CIs include true values
    irf_include_wbr68[:, :, sim] \
        = check_include(K, n_imp, WBR_results["ci68_irf"], IRF_true)
    irf_norm_include_wbr68[:, :, sim] \
        = check_include(K, n_imp, WBR_results["ci68_irf_norm"], IRFnorm_true)
    fevd_include_wbr68[:, :, sim] \
        = check_include(K, n_imp, WBR_results["ci68_fevd"], FEVD_true)
    irf_include_wbr95[:, :, sim] \
        = check_include(K, n_imp, WBR_results["ci95_irf"], IRF_true)
    irf_norm_include_wbr95[:, :, sim] \
        = check_include(K, n_imp, WBR_results["ci95_irf_norm"], IRFnorm_true)
    fevd_include_wbr95[:, :, sim] \
        = check_include(K, n_imp, WBR_results["ci95_fevd"], FEVD_true)
    
    # Check if moving block bootstrap Hall's intervals include true values
    irf_include_halls68[:, :, sim] \
        = check_include(K, n_imp, ci68halls_irf, IRF_true)
    irf_norm_include_halls68[:, :, sim] \
        = check_include(K, n_imp, ci68halls_irf_norm, IRFnorm_true)
    fevd_include_halls68[:, :, sim] \
        = check_include(K, n_imp, ci68halls_fevd, FEVD_true)
    irf_include_halls95[:, :, sim] \
        = check_include(K, n_imp, ci95halls_irf, IRF_true)
    irf_norm_include_halls95[:, :, sim] \
        = check_include(K, n_imp, ci95halls_irf_norm, IRFnorm_true)
    fevd_include_halls95[:, :, sim] \
        = check_include(K, n_imp, ci95halls_fevd, FEVD_true)
    
    # Check if normalized IRFs are included in the AR grid confidence sets
    grid = ARSim_results["ar_grid"]
    ar_sim_set68 = ARSim_results["ar_index68"]
    ar_sim_set95 = ARSim_results["ar_index95"]
    counter = 0
    for kk in range(K):
        for ii in range(n_imp):
            ar_sim_include68[kk, ii, sim] = ar_sim_set68[counter, ii, kk]
            ar_sim_include95[kk, ii, sim] = ar_sim_set95[counter, ii, kk]
            counter = counter + 1
    
    # Check if extreme values are included in AR grid confidence sets
    for kk in range(K):
        for ii in range(n_imp):
            ar_extreme_sim_include68[kk, ii, sim] \
                = np.all([ar_sim_set68[2*n_imp, ii, kk],
                          ar_sim_set68[2*n_imp+1, ii, kk]])
            ar_extreme_sim_include95[kk, ii, sim] \
                = np.all([ar_sim_set95[2*n_imp, ii, kk],
                          ar_sim_set95[2*n_imp+1, ii, kk]])
    
    # Store the tests of proxy strength
    f_store[sim] = SVAR_results['f_stat']
    f_iv_store[sim] = SVAR_results['f_iv_stat']
    w_store[sim] = MSW_results['wald_stat']
    
    # Compute the length of confidence intervals/sets
    if chooseDGP == 1:
        # Compute the length of the MBB percentile intervals
        ci68_irf_norm = MBB_results["ci68_irf_norm"]
        len_mbb68[:, :, sim] = (ci68_irf_norm[1, :, :] - ci68_irf_norm[0, :, :]).T
        ci95_irf_norm = MBB_results["ci95_irf_norm"]
        len_mbb95[:, :, sim] = (ci95_irf_norm[1, :, :] - ci95_irf_norm[0, :, :]).T
        
        # Compute the length of the AR-type grid confidence sets
        ar_MBBindex68 = MBB_results["ar_index68"]
        ar_MBBindex95 = MBB_results["ar_index95"]
        for kk in range(K):
            for ii in range(n_imp):
                # AR-type grid sets
                mask68 = ar_MBBindex68[:, ii, kk] == 1
                ar_points = ARgrid[mask68]
                if ar_points.size > 0:
                    len_mbb_ar68[kk, ii, sim] = ar_points.max() - ar_points.min()
                else:
                    len_mbb_ar68[kk, ii, sim] = 0
                
                mask95 = ar_MBBindex95[:, ii, kk] == 1
                ar_points = ARgrid[mask95]
                if ar_points.size > 0:
                    len_mbb_ar95[kk, ii, sim] = ar_points.max() - ar_points.min()
                else:
                    len_mbb_ar95[kk, ii, sim] = 0
        
        # Compute the length of the MSW analytic confidence set
        len_msw68[:, :, sim] = (cs68_msw[1, :, :] - cs68_msw[0, :, :]).T
        len_msw95[:, :, sim] = (cs95_msw[1, :, :] - cs95_msw[0, :, :]).T
        
        # Compute the length of the WBR percentile intervals
        ci68wbr_irf_norm = WBR_results["ci68_irf_norm"]
        len_wbr68[:, :, sim] \
            = (ci68wbr_irf_norm[1, :, :] - ci68wbr_irf_norm[0, :, :]).T
        ci95wbr_irf_norm = WBR_results["ci95_irf_norm"]
        len_wbr95[:, :, sim] \
            = (ci95wbr_irf_norm[1, :, :] - ci95wbr_irf_norm[0, :, :]).T
        
        # Compute the length of the MBB Hall's intervals
        len_mbb_halls68[:, :, sim] \
            = (ci68halls_irf_norm[1, :, :] - ci68halls_irf_norm[0, :, :]).T
        len_mbb_halls95[:, :, sim] \
            = (ci95halls_irf_norm[1, :, :] - ci95halls_irf_norm[0, :, :]).T
        
        # Compute the length of the AR-type grid confidence sets
        ar_SIMindex68 = ARSim_results["ar_index68"]
        ar_SIMindex95 = ARSim_results["ar_index95"]
        for kk in range(K):
            for ii in range(n_imp):
                # AR-type grid sets
                mask68 = ar_SIMindex68[:, ii, kk] == 1
                mask95 = ar_SIMindex95[:, ii, kk] == 1
                ar_points = ARgrid[mask68]
                if ar_points.size > 0:
                    len_sim_ar68[kk, ii, sim] = ar_points.max() - ar_points.min()
                else:
                    len_sim_ar68[kk, ii, sim] = 0
                
                ar_points = ARgrid[mask95]
                if ar_points.size > 0:
                    len_sim_ar95[kk, ii, sim] = ar_points.max() - ar_points.min()
                else:
                    len_sim_ar95[kk, ii, sim] = 0
    
    # Show the loop
    print(f'You finished simulation loop number {sim+1} out of {nSim}.')


#%% Show the Results for the Tests of Weak Proxies

# Check if any extreme value is included in the AR MBB set
extreme_store68 = np.zeros(nSim)
extreme_store95 = np.zeros(nSim)
for sim in range(nSim):
    extreme_store68[sim] = np.any(ar_extreme_include68[:, :, sim])
    extreme_store95[sim] = np.any(ar_extreme_include95[:, :, sim])

# Show results
print(' ')
print('For sample size ' + str(n_sample) + ':')

# Results for the F statistic
print('The F stat was above 10 in '
      + str(100*np.mean((f_store > 10))) + ' percent of simulations.')

# Results for the F_iv statistic
print('The F_iv stat was above 10 in '
      + str(100*np.mean((f_iv_store > 10))) + ' percent of simulations.')

# Results for the Wald statistic
print('The Wald stat was below 0.9889 in '
      + str(100*np.mean((f_iv_store < 0.9889))) + ' percent of simulations.')

# Results for the Wald statistic
print('The Wald stat was below 3.8415 in '
      + str(100*np.mean((f_iv_store < 3.8415))) + ' percent of simulations.')

# Results for the AR MBB set
print('The 68% MBB AR-type set included an extreme value in '
      + str(100*np.mean(extreme_store68)) + ' percent of simulations.')

# Results for the AR MBB set
print('The 95% MBB AR-type set included an extreme value in '
      + str(100*np.mean(extreme_store95)) + ' percent of simulations.')


#%% Compute the Coverage Rates of One Standard Deviation IRFs

# MBB percentile coverage rates
IRF_cover68 = np.zeros((K, n_imp))
IRF_cover68[:, :] = np.mean(irf_include68, axis=2)
IRF_cover95 = np.zeros((K, n_imp))
IRF_cover95[:, :] = np.mean(irf_include95, axis=2)

# Wild bootstrap percentile coverage rates
IRF_cover_wbr68 = np.zeros((K, n_imp))
IRF_cover_wbr68[:, :] = np.mean(irf_include_wbr68, axis=2)
IRF_cover_wbr95 = np.zeros((K, n_imp))
IRF_cover_wbr95[:, :] = np.mean(irf_include_wbr95, axis=2)

# MBB Hall's percentile coverage rates
IRF_cover_halls68 = np.zeros((K, n_imp))
IRF_cover_halls68[:, :] = np.mean(irf_include_halls68, axis=2)
IRF_cover_halls95 = np.zeros((K, n_imp))
IRF_cover_halls95[:, :] = np.mean(irf_include_halls95, axis=2)

# Store in a pandas dataframe
cols_irf = ['irf_mbb1', 'irf_mbb2', 'irf_wbr1', 'irf_wbr2',
            'irf_mbb_halls1', 'irf_mbb_halls2']
df_irf68 = pd.DataFrame(index=np.arange(0, n_imp), columns=cols_irf)
df_irf95 = pd.DataFrame(index=np.arange(0, n_imp), columns=cols_irf)

df_irf68.loc[:, 'irf_mbb1'] = IRF_cover68[0, :]
df_irf68.loc[:, 'irf_mbb2'] = IRF_cover68[1, :]
df_irf68.loc[:, 'irf_wbr1'] = IRF_cover_wbr68[0, :]
df_irf68.loc[:, 'irf_wbr2'] = IRF_cover_wbr68[1, :]
df_irf68.loc[:, 'irf_mbb_halls1'] = IRF_cover_halls68[0, :]
df_irf68.loc[:, 'irf_mbb_halls2'] = IRF_cover_halls68[1, :]

df_irf95.loc[:, 'irf_mbb1'] = IRF_cover95[0, :]
df_irf95.loc[:, 'irf_mbb2'] = IRF_cover95[1, :]
df_irf95.loc[:, 'irf_wbr1'] = IRF_cover_wbr95[0, :]
df_irf95.loc[:, 'irf_wbr2'] = IRF_cover_wbr95[1, :]
df_irf95.loc[:, 'irf_mbb_halls1'] = IRF_cover_halls95[0, :]
df_irf95.loc[:, 'irf_mbb_halls2'] = IRF_cover_halls95[1, :]


#%% Compute the Coverage Rates of Normalized IRFs

# MBB percentile coverage rates
IRFnorm_cover68 = np.zeros((K, n_imp))
IRFnorm_cover68[:, :] = np.mean(irf_norm_include68, axis=2)
IRFnorm_cover95 = np.zeros((K, n_imp))
IRFnorm_cover95[:, :] = np.mean(irf_norm_include95, axis=2)

# Anderson Rubin type grid coverage rates
AR_mbb_cover68 = np.zeros((K, n_imp))
AR_mbb_cover68[:, :] = np.mean(ar_mbb_include68, axis=2)
AR_mbb_cover95 = np.zeros((K, n_imp))
AR_mbb_cover95[:, :] = np.mean(ar_mbb_include95, axis=2)

# Montiel Olea, Stock and Watson coverage rates
MSW_cover68 = np.zeros((K, n_imp))
MSW_cover68[:, :] = np.mean(msw_include68, axis=2)
MSW_cover95 = np.zeros((K, n_imp))
MSW_cover95[:, :] = np.mean(msw_include95, axis=2)

# Wild bootstrap percentile coverage rates
IRFnorm_cover_wbr68 = np.zeros((K, n_imp))
IRFnorm_cover_wbr68[:, :] = np.mean(irf_norm_include_wbr68, axis=2)
IRFnorm_cover_wbr95 = np.zeros((K, n_imp))
IRFnorm_cover_wbr95[:, :] = np.mean(irf_norm_include_wbr95, axis=2)

# MBB Hall's percentile coverage rates
IRFnorm_cover_halls68 = np.zeros((K, n_imp))
IRFnorm_cover_halls68[:, :] = np.mean(irf_norm_include_halls68, axis=2)
IRFnorm_cover_halls95 = np.zeros((K, n_imp))
IRFnorm_cover_halls95[:, :] = np.mean(irf_norm_include_halls95, axis=2)

# Anderson Rubin type grid coverage rates
AR_sim_cover68 = np.zeros((K, n_imp))
AR_sim_cover68[:, :] = np.mean(ar_sim_include68, axis=2)
AR_sim_cover95 = np.zeros((K, n_imp))
AR_sim_cover95[:, :] = np.mean(ar_sim_include95, axis=2)

# Store in a pandas dataframe
cols_irfnorm = ['irfnorm_mbb1', 'irfnorm_mbb2', 'ar_mbb1', 'ar_mbb2',
                'msw1', 'msw2', 'irfnorm_wbr1', 'irfnorm_wbr2',
                'irfnorm_mbb_halls1', 'irfnorm_mbb_halls2',
                'ar_sim1', 'ar_sim2']
df_irfnorm68 = pd.DataFrame(index=np.arange(0, n_imp), columns=cols_irfnorm)
df_irfnorm95 = pd.DataFrame(index=np.arange(0, n_imp), columns=cols_irfnorm)

df_irfnorm68.loc[:, 'irfnorm_mbb1'] = IRFnorm_cover68[0, :]
df_irfnorm68.loc[:, 'irfnorm_mbb2'] = IRFnorm_cover68[1, :]
df_irfnorm68.loc[:, 'ar_mbb1'] = AR_mbb_cover68[0, :]
df_irfnorm68.loc[:, 'ar_mbb2'] = AR_mbb_cover68[1, :]
df_irfnorm68.loc[:, 'msw1'] = MSW_cover68[0, :]
df_irfnorm68.loc[:, 'msw2'] = MSW_cover68[1, :]
df_irfnorm68.loc[:, 'irfnorm_wbr1'] = IRFnorm_cover_wbr68[0, :]
df_irfnorm68.loc[:, 'irfnorm_wbr2'] = IRFnorm_cover_wbr68[1, :]
df_irfnorm68.loc[:, 'irfnorm_mbb_halls1'] = IRFnorm_cover_halls68[0, :]
df_irfnorm68.loc[:, 'irfnorm_mbb_halls2'] = IRFnorm_cover_halls68[1, :]
df_irfnorm68.loc[:, 'ar_sim1'] = AR_sim_cover68[0, :]
df_irfnorm68.loc[:, 'ar_sim2'] = AR_sim_cover68[1, :]

df_irfnorm95.loc[:, 'irfnorm_mbb1'] = IRFnorm_cover95[0, :]
df_irfnorm95.loc[:, 'irfnorm_mbb2'] = IRFnorm_cover95[1, :]
df_irfnorm95.loc[:, 'ar_mbb1'] = AR_mbb_cover95[0, :]
df_irfnorm95.loc[:, 'ar_mbb2'] = AR_mbb_cover95[1, :]
df_irfnorm95.loc[:, 'msw1'] = MSW_cover95[0, :]
df_irfnorm95.loc[:, 'msw2'] = MSW_cover95[1, :]
df_irfnorm95.loc[:, 'irfnorm_wbr1'] = IRFnorm_cover_wbr95[0, :]
df_irfnorm95.loc[:, 'irfnorm_wbr2'] = IRFnorm_cover_wbr95[1, :]
df_irfnorm95.loc[:, 'irfnorm_mbb_halls1'] = IRFnorm_cover_halls95[0, :]
df_irfnorm95.loc[:, 'irfnorm_mbb_halls2'] = IRFnorm_cover_halls95[1, :]
df_irfnorm95.loc[:, 'ar_sim1'] = AR_sim_cover95[0, :]
df_irfnorm95.loc[:, 'ar_sim2'] = AR_sim_cover95[1, :]


#%% Compute the Coverage Rates of FEVDs

# MBB percentile coverage rates for FEVDs
FEVD_cover68 = np.zeros((K, n_imp))
FEVD_cover68[:, :] = np.mean(fevd_include68, axis=2)
FEVD_cover95 = np.zeros((K, n_imp))
FEVD_cover95[:, :] = np.mean(fevd_include95, axis=2)

# Wild bootstrap percentile coverage rates
FEVD_cover_wbr68 = np.zeros((K, n_imp))
FEVD_cover_wbr68[:, :] = np.mean(fevd_include_wbr68, axis=2)
FEVD_cover_wbr95 = np.zeros((K, n_imp))
FEVD_cover_wbr95[:, :] = np.mean(fevd_include_wbr95, axis=2)

# MBB Hall's percentile coverage rates
FEVD_cover_halls68 = np.zeros((K, n_imp))
FEVD_cover_halls68[:, :] = np.mean(fevd_include_halls68, axis=2)
FEVD_cover_halls95 = np.zeros((K, n_imp))
FEVD_cover_halls95[:, :] = np.mean(fevd_include_halls95, axis=2)

# Store in a pandas dataframe
cols_fevd = ['fevd_mbb1', 'fevd_mbb2', 'fevd_wbr1', 'fevd_wbr2',
             'fevd_mbb_halls1', 'fevd_mbb_halls2']
df_fevd68 = pd.DataFrame(index=np.arange(0, n_imp), columns=cols_fevd)
df_fevd95 = pd.DataFrame(index=np.arange(0, n_imp), columns=cols_fevd)

df_fevd68.loc[:, 'fevd_mbb1'] = FEVD_cover68[0, :]
df_fevd68.loc[:, 'fevd_mbb2'] = FEVD_cover68[1, :]
df_fevd68.loc[:, 'fevd_wbr1'] = FEVD_cover_wbr68[0, :]
df_fevd68.loc[:, 'fevd_wbr2'] = FEVD_cover_wbr68[1, :]
df_fevd68.loc[:, 'fevd_mbb_halls1'] = FEVD_cover_halls68[0, :]
df_fevd68.loc[:, 'fevd_mbb_halls2'] = FEVD_cover_halls68[1, :]

df_fevd95.loc[:, 'fevd_mbb1'] = FEVD_cover95[0, :]
df_fevd95.loc[:, 'fevd_mbb2'] = FEVD_cover95[1, :]
df_fevd95.loc[:, 'fevd_wbr1'] = FEVD_cover_wbr95[0, :]
df_fevd95.loc[:, 'fevd_wbr2'] = FEVD_cover_wbr95[1, :]
df_fevd95.loc[:, 'fevd_mbb_halls1'] = FEVD_cover_halls95[0, :]
df_fevd95.loc[:, 'fevd_mbb_halls2'] = FEVD_cover_halls95[1, :]


#%% Compute the Avg Lengths of Confidence Intervals/Sets for Normalized IRFs

if chooseDGP == 1:
    # Computations
    len_mbb_avg68 = np.nanmean(len_mbb68, axis=2)
    len_mbb_ar_avg68 = np.nanmean(len_mbb_ar68, axis=2)
    len_msw_avg68 = np.nanmean(len_msw68, axis=2)
    len_wbr_avg68 = np.nanmean(len_wbr68, axis=2)
    len_mbb_halls_avg68 = np.nanmean(len_mbb_halls68, axis=2)
    len_sim_ar_avg68 = np.nanmean(len_sim_ar68, axis=2)
    
    len_mbb_avg95 = np.nanmean(len_mbb95, axis=2)
    len_mbb_ar_avg95 = np.nanmean(len_mbb_ar95, axis=2)
    len_msw_avg95 = np.nanmean(len_msw95, axis=2)
    len_wbr_avg95 = np.nanmean(len_wbr95, axis=2)
    len_mbb_halls_avg95 = np.nanmean(len_mbb_halls95, axis=2)
    len_sim_ar_avg95 = np.nanmean(len_sim_ar95, axis=2)
    
    # Store in a pandas dataframe
    cols_length = ['irfnorm_mbb1', 'irfnorm_mbb2', 'ar_mbb1', 'ar_mbb2',
                   'msw1', 'msw2', 'irfnorm_wbr1', 'irfnorm_wbr2',
                   'irfnorm_mbb_halls1', 'irfnorm_mbb_halls2',
                   'ar_sim1', 'ar_sim2']
    df_length68 = pd.DataFrame(index=np.arange(0, n_imp), columns=cols_length)
    df_length95 = pd.DataFrame(index=np.arange(0, n_imp), columns=cols_length)
    
    df_length68.loc[:, 'irfnorm_mbb1'] = len_mbb_avg68[0, :]
    df_length68.loc[:, 'irfnorm_mbb2'] = len_mbb_avg68[1, :]
    df_length68.loc[:, 'ar_mbb1'] = len_mbb_ar_avg68[0, :]
    df_length68.loc[:, 'ar_mbb2'] = len_mbb_ar_avg68[1, :]
    df_length68.loc[:, 'msw1'] = len_msw_avg68[0, :]
    df_length68.loc[:, 'msw2'] = len_msw_avg68[1, :]
    df_length68.loc[:, 'irfnorm_wbr1'] = len_wbr_avg68[0, :]
    df_length68.loc[:, 'irfnorm_wbr2'] = len_wbr_avg68[1, :]
    df_length68.loc[:, 'irfnorm_mbb_halls1'] = len_mbb_halls_avg68[0, :]
    df_length68.loc[:, 'irfnorm_mbb_halls2'] = len_mbb_halls_avg68[1, :]
    df_length68.loc[:, 'ar_sim1'] = len_sim_ar_avg68[0, :]
    df_length68.loc[:, 'ar_sim2'] = len_sim_ar_avg68[1, :]
    
    df_length95.loc[:, 'irfnorm_mbb1'] = len_mbb_avg95[0, :]
    df_length95.loc[:, 'irfnorm_mbb2'] = len_mbb_avg95[1, :]
    df_length95.loc[:, 'ar_mbb1'] = len_mbb_ar_avg95[0, :]
    df_length95.loc[:, 'ar_mbb2'] = len_mbb_ar_avg95[1, :]
    df_length95.loc[:, 'msw1'] = len_msw_avg95[0, :]
    df_length95.loc[:, 'msw2'] = len_msw_avg95[1, :]
    df_length95.loc[:, 'irfnorm_wbr1'] = len_wbr_avg95[0, :]
    df_length95.loc[:, 'irfnorm_wbr2'] = len_wbr_avg95[1, :]
    df_length95.loc[:, 'irfnorm_mbb_halls1'] = len_mbb_halls_avg95[0, :]
    df_length95.loc[:, 'irfnorm_mbb_halls2'] = len_mbb_halls_avg95[1, :]
    df_length95.loc[:, 'ar_sim1'] = len_sim_ar_avg95[0, :]
    df_length95.loc[:, 'ar_sim2'] = len_sim_ar_avg95[1, :]


#%% Save the Results

# Save to Excel
str_file68 = 'dgp' + str(chooseDGP) + '_results68.xlsx'
str_file95 = 'dgp' + str(chooseDGP) + '_results95.xlsx'

if chooseDGP == 1:
    with pd.ExcelWriter(str_file68) as writer:
        df_parm.to_excel(writer, sheet_name='parameters')
        df_irf68.to_excel(writer, sheet_name='irf')
        df_irfnorm68.to_excel(writer, sheet_name='irfnorm')
        df_fevd68.to_excel(writer, sheet_name='fevd')
        df_length68.to_excel(writer, sheet_name='length')
    
    with pd.ExcelWriter(str_file95) as writer:
        df_parm.to_excel(writer, sheet_name='parameters')
        df_irf95.to_excel(writer, sheet_name='irf')
        df_irfnorm95.to_excel(writer, sheet_name='irfnorm')
        df_fevd95.to_excel(writer, sheet_name='fevd')
        df_length95.to_excel(writer, sheet_name='length')
else:
    with pd.ExcelWriter(str_file68) as writer:
        df_parm.to_excel(writer, sheet_name='parameters')
        df_irf68.to_excel(writer, sheet_name='irf')
        df_irfnorm68.to_excel(writer, sheet_name='irfnorm')
        df_fevd68.to_excel(writer, sheet_name='fevd')
    
    with pd.ExcelWriter(str_file95) as writer:
        df_parm.to_excel(writer, sheet_name='parameters')
        df_irf95.to_excel(writer, sheet_name='irf')
        df_irfnorm95.to_excel(writer, sheet_name='irfnorm')
        df_fevd95.to_excel(writer, sheet_name='fevd')


#%% Disply the Run Time
print(' ')
print('This script finished in ' 
      + str(np.round((time.time()-start)/60, decimals = 1)) + ' minutes.')
