# -*- coding: utf-8 -*-
"""
This script produces results shown in Appendix G.8 of "Asymptotically Valid
Bootstrap Inference for Proxy SVARs" by Carsten Jentsch and Kurt G. Lunsford.

The results are automatically saved in the following Excel file:
    lag_augmented_results.xlsx
When this file is saved, it is called by other scripts to produce figures in
Appendix G.8.

This script uses functions from the following modules:
    proxy_svar_module.py
    proxy_svar_aug_module.py
    utilities_module.py

This script also uses numpy, pandas, os, sys, and time packages.
"""

#%% Import Packages and Functions

# Pandas and NumPy
import pandas as pd
import numpy as np

# Get replication files directory
import os
path_dir = os.path.dirname(os.getcwd())

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Functions for appendix simulations
from modules.proxy_svar_module import make_dynamics, make_proxy_svar_dynamics
from modules.proxy_svar_module import make_proxy_svar_mbb

from modules.proxy_svar_aug_module import make_proxy_svar_dynamics_aug
from modules.proxy_svar_aug_module import make_proxy_svar_mbb_aug

from modules.utilities_module import make_data_sv, check_include

# Time the program
import time
start = time.time()


#%% Parameters of the Script

# Parameters of the SVAR
A0 = np.zeros((2, 1))
A1 = np.array([(0.44, 0.66), (-0.11, 1.32)])
A2 = np.array([(-0.18, 0), (-0.18, -0.09)])
A = np.hstack((A0, A1, A2))
H = np.array([(np.sqrt(2)/2, np.sqrt(2)/2),
              ((np.sqrt(2)-np.sqrt(6))/4, (np.sqrt(2)+np.sqrt(6))/4)])

# Parameters of stochastic volatility
rho = 0.85
sigma_e = 0.15
log_sig_bar = -sigma_e*sigma_e/(1 - rho*rho)

# Covariance between structural shock of interest and proxy variable
psi = 1.0

# Population mean of the proxy variable
m_bar = 0

# Fraction of proxy variables that are censored
exp_d = 0

# IRF and FEVD parameters
s = -1          # scale for the normalized IRF
n_imp = 21      # horizon for IRFs and FEVDs, including the impact (0 horizon)

# Simulation parameters
n_sample = 200                  # sample sizes to be used
burn = 1000                     # number of observations to be burned
blocksize = 4                   # block size for each sample size to be used
nBoot = 2000                    # number of bootstrap replications
nSim = 1000                     # number of simulations

# Seed the random number generator
np.random.seed(0)


#%% Store the Parameters in a DataFrame

cols_parm = ['n_sample', 'ell', 'psi']
df_parm = pd.DataFrame(index=[0], columns=cols_parm)

df_parm.loc[:, 'n_sample'] = n_sample
df_parm.loc[:, 'ell'] = blocksize
df_parm.loc[:, 'psi'] = psi


#%% Compute Additional Parameters and the Population IRFs and FEVDs

# Compute dimensions of the data generating process
K = A[:, 0].size                # Number of variables in the VAR
p = int((A[0, :].size - 1)/K)   # Number of lags of the VAR

# Compute some population parameters
H1 = H[:, 0:1]                  # First column of H
SigmaU = H @ H.T                # Covariance matrix of VAR innovations

# Compute IRFs and FEVDs
IRF_true, IRFnorm_true, FEVD_true, SVMA_true \
    = make_dynamics(A, SigmaU, H1, H1, p, s, n_imp)


#%% Set the Grid for the Anderson Rubin Grid Bootstrap

# Construct the grid include the values in the true IRF
ARgrid = np.empty(K*n_imp)
ARgrid[:] = IRFnorm_true.flatten()

# Add extreme values to the grid check if the confidence set is "ubounded"
ARgrid = np.append(ARgrid, np.array([-1000, 1000]))

# Add additional points to get a sense of the length of the confidence set
temp = np.abs(s)*np.arange(-16, 44, 0.02)
ARgrid = np.append(ARgrid, temp)


#%% Storage Arrays for Simulation Results

# Storage arrays for moving block bootstrap percentile intervals
irf_include68 = np.zeros((K, n_imp, nSim))
irf_norm_include68 = np.zeros((K, n_imp, nSim))
fevd_include68 = np.zeros((K, n_imp, nSim))

# Storage arrays for Anderson Rubin type grid confidence set
ar_mbb_include68 = np.zeros((K, n_imp, nSim))
ar_extreme_include68 = np.zeros((K, n_imp, nSim))

# Storage arrays for lag-augmented moving block bootstrap percentile intervals
irf_include_aug68 = np.zeros((K, n_imp, nSim))
irf_norm_include_aug68 = np.zeros((K, n_imp, nSim))
fevd_include_aug68 = np.zeros((K, n_imp, nSim))

# Storage arrays for lag-augmented Anderson Rubin type grid confidence set
ar_mbb_include_aug68 = np.zeros((K, n_imp, nSim))
ar_extreme_include_aug68 = np.zeros((K, n_imp, nSim))

# Storage arrays for F statistics
f_store = np.zeros(nSim)
f_iv_store = np.zeros(nSim)

# Storage arrays to measure the length confidence intervals/sets
len_mbb = np.zeros((K, n_imp, nSim))
len_mbb_aug = np.zeros((K, n_imp, nSim))
len_mbb_ar = np.zeros((K, n_imp, nSim))
len_mbb_ar_aug = np.zeros((K, n_imp, nSim))
len_mbb_fevd = np.zeros((K, n_imp, nSim))
len_mbb_fevd_aug = np.zeros((K, n_imp, nSim))


#%% Simulation

for sim in range(nSim):
    # Construct the observable time series
    y, shocks, vv = make_data_sv((burn+np.max(n_sample)), K, p, A, H,
                                 rho, log_sig_bar, sigma_e)
    
    # Drop the burn period less the initial p observations of VAR data
    y = y[burn-p:, :]
    
    # Drop the burn period of the proxy variable components
    shocks = shocks[burn-p:, :]
    vv = vv[burn-p:, :]
    
    # Compute the censoring
    d = (np.random.uniform(size=((n_sample+p), 1)) > exp_d)
    
    # Set the data sample
    var_data = np.zeros((n_sample+p, K))
    var_data = y[0:n_sample+p, :]
    
    # Construct the proxy variable
    proxies = np.zeros((n_sample+p, 1))
    proxies[:, :] = m_bar + psi*shocks[0:n_sample+p, 0:1] + vv[0:n_sample+p, :]
    proxies[:, :] = np.multiply(proxies, d[0:n_sample+p, :])
    
    # Estimate the proxy SVAR and compute the dynamics
    SVAR_results = make_proxy_svar_dynamics(var_data, proxies, p, s, n_imp)
    
    # Save some estimates for additional calculations
    a_hat = np.zeros((K, 1+K*p))
    a_hat[:, :] = SVAR_results['a_hat']
    
    u_hat = np.zeros((n_sample, K))
    u_hat[:, :] = SVAR_results['u_hat']
    
    svma_hat = np.zeros((K, n_imp))
    svma_hat = SVAR_results['svma_hat']
    
    # Set the initial conditions for the bootstrap without lag augmenting
    y0 = np.zeros(1+K*p)
    y0[0] = 1
    for ii in range(p):
        y0[1+p*ii:1+p*(ii+1)] = var_data[p-1-ii, :]
    
    # Truncate the proxy variables to align with the VAR residuals
    proxies_trunc = np.zeros((n_sample, 1))
    proxies_trunc[:, 0] = proxies[p:n_sample+p, 0]
    
    # Compute moving block bootstrap CIs
    MBB_results = make_proxy_svar_mbb(a_hat, u_hat, proxies_trunc, y0,
                                      p, s, n_imp, nBoot, blocksize,
                                      toggleAR=1, ar_grid=ARgrid,
                                      svma_vals=svma_hat)
    
    # Estimate the lag-augmented proxy SVAR and compute the dynamics
    SVAR_results_aug \
        = make_proxy_svar_dynamics_aug(var_data, proxies, p, s, n_imp)
    
    # Save some estimates for additional calculations
    a_aug = np.zeros((K, 1+K*(p+1)))
    a_aug[:, :] = SVAR_results_aug['a_hat']
    
    u_aug = np.zeros((n_sample-1, K))
    u_aug[:, :] = SVAR_results_aug['u_hat']
    
    svma_aug = np.zeros((K, n_imp))
    svma_aug = SVAR_results_aug['svma_hat']
    
    # Set the initial conditions for the bootstrap with lag augmenting
    y0 = np.zeros(1+K*(p+1))
    y0[0] = 1
    for ii in range(p+1):
        y0[1+p*ii:1+p*(ii+1)] = var_data[p-ii, :]
    
    # Truncate the proxy variables to align with the lag-augmented VAR residuals
    proxies_trunc = np.zeros((n_sample-1, 1))
    proxies_trunc[:, 0] = proxies[p+1:n_sample+p, 0]
    
    # Compute moving block bootstrap CIs
    MBB_results_aug = make_proxy_svar_mbb_aug(a_aug, u_aug, proxies_trunc, y0,
                                              p, s, n_imp, nBoot, blocksize,
                                              toggleAR=1, ar_grid=ARgrid,
                                              svma_vals=svma_aug)
    
    # Check if moving block bootstrap CIs include true values
    irf_include68[:, :, sim] \
        = check_include(K, n_imp, MBB_results["ci68_irf"], IRF_true)
    irf_norm_include68[:, :, sim] \
        = check_include(K, n_imp, MBB_results["ci68_irf_norm"], IRFnorm_true)
    fevd_include68[:, :, sim] \
        = check_include(K, n_imp, MBB_results["ci68_fevd"], FEVD_true)
    
    # Check if lag augmented moving block bootstrap CIs include true values
    irf_include_aug68[:, :, sim] \
        = check_include(K, n_imp, MBB_results_aug["ci68_irf"], IRF_true)
    irf_norm_include_aug68[:, :, sim] \
        = check_include(K, n_imp, MBB_results_aug["ci68_irf_norm"], IRFnorm_true)
    fevd_include_aug68[:, :, sim] \
        = check_include(K, n_imp, MBB_results_aug["ci68_fevd"], FEVD_true)
    
    # Check if normalized IRFs are included in the AR grid confidence sets
    grid = MBB_results["ar_grid"]
    ar_mbb_set68 = MBB_results["ar_index68"]
    counter = 0
    for kk in range(K):
        for ii in range(n_imp):
            ar_mbb_include68[kk, ii, sim] = ar_mbb_set68[counter, ii, kk]
            counter = counter + 1
    
    # Check if normalized IRFs are included in the lag augmented AR grid confidence sets
    grid = MBB_results_aug["ar_grid"]
    ar_mbb_set_aug68 = MBB_results_aug["ar_index68"]
    counter = 0
    for kk in range(K):
        for ii in range(n_imp):
            ar_mbb_include_aug68[kk, ii, sim] = ar_mbb_set_aug68[counter, ii, kk]
            counter = counter + 1
    
    # Check if extreme values are included in AR grid confidence sets
    temp = np.zeros
    for kk in range(K):
        for ii in range(n_imp):
            ar_extreme_include68[kk, ii, sim] \
                = np.all([ar_mbb_set68[2*n_imp, ii, kk],
                          ar_mbb_set68[2*n_imp+1, ii, kk]])
    
    # Check if extreme values are included in lag augmented AR grid confidence sets
    temp = np.zeros
    for kk in range(K):
        for ii in range(n_imp):
            ar_extreme_include_aug68[kk, ii, sim] \
                = np.all([ar_mbb_set_aug68[2*n_imp, ii, kk],
                          ar_mbb_set_aug68[2*n_imp+1, ii, kk]])
    
    # Compute the length of the MBB percentile intervals
    ci68_irf_norm = MBB_results["ci68_irf_norm"]
    len_mbb[:, :, sim] = (ci68_irf_norm[1, :, :] - ci68_irf_norm[0, :, :]).T
    
    # Compute the length of the lag augmented MBB percentile intervals
    ci68_irf_norm_aug = MBB_results_aug["ci68_irf_norm"]
    len_mbb_aug[:, :, sim] \
        = (ci68_irf_norm_aug[1, :, :] - ci68_irf_norm_aug[0, :, :]).T
    
    # Compute the length of the AR-type grid confidence sets
    ar_MBBindex68 = MBB_results["ar_index68"]
    ar_MBBindex_aug68 = MBB_results_aug["ar_index68"]
    for kk in range(K):
        for ii in range(n_imp):
            # AR-type grid sets
            mask = ar_MBBindex68[:, ii, kk] == 1
            ar_points = ARgrid[mask]
            if ar_points.size > 0:
                len_mbb_ar[kk, ii, sim] = ar_points.max() - ar_points.min()
            else:
                len_mbb_ar[kk, ii, sim] = 0
            
            mask = ar_MBBindex_aug68[:, ii, kk] == 1
            ar_points = ARgrid[mask]
            if ar_points.size > 0:
                len_mbb_ar_aug[kk, ii, sim] = ar_points.max() - ar_points.min()
            else:
                len_mbb_ar_aug[kk, ii, sim] = 0
    
    # Compute the length of the MBB percentile intervals
    ci68_fevd = MBB_results["ci68_fevd"]
    len_mbb_fevd[:, :, sim] = (ci68_fevd[1, :, :] - ci68_fevd[0, :, :]).T
    
    # Compute the length of the lag augmented MBB percentile intervals
    ci68_fevd_aug = MBB_results_aug["ci68_fevd"]
    len_mbb_fevd_aug[:, :, sim] \
        = (ci68_fevd_aug[1, :, :] - ci68_fevd_aug[0, :, :]).T
    
    # Store the tests of proxy strength
    f_store[sim] = SVAR_results['f_stat']
    f_iv_store[sim] = SVAR_results['f_iv_stat']
    
    # Show the loop
    print(f'You finished simulation loop number {sim+1} out of {nSim}.')


#%% Show the Results for the Tests of Weak Proxies

# Check if any extreme value is included in the AR MBB set
extreme_store = np.zeros(nSim)
for sim in range(nSim):
    extreme_store[sim] = np.any(ar_extreme_include68[:, :, sim])

# Show results
print(' ')
print('For sample size ' + str(n_sample) + ':')

# Results for the F statistic
print('The F stat was above 10 in '
      + str(100*np.mean((f_store > 10))) + ' percent of simulations.')

# Results for the F_iv statistic
print('The F_iv stat was above 10 in '
      + str(100*np.mean((f_iv_store > 10))) + ' percent of simulations.')

# Results for the Wald statistic
print('The Wald stat was below 0.9889 in '
      + str(100*np.mean((f_iv_store < 0.9889))) + ' percent of simulations.')

# Results for the AR MBB set
print('The AR MBB set included an extreme value in '
      + str(100*np.mean(extreme_store)) + ' percent of simulations.')


#%% Compute the Coverage Rates of One Standard Deviation IRFs

# MBB percentile coverage rates
IRF_cover68 = np.zeros((K, n_imp))
IRF_cover68[:, :] = np.mean(irf_include68, axis=2)

# Lag augmented MBB percentile coverage rates
IRF_cover_aug68 = np.zeros((K, n_imp))
IRF_cover_aug68[:, :] = np.mean(irf_include_aug68, axis=2)

# Store in a pandas dataframe
cols_irf = ['irf_mbb1', 'irf_mbb2', 'irf_mbb_aug1', 'irf_mbb_aug2']
df_irf = pd.DataFrame(index=np.arange(0, n_imp), columns=cols_irf)

df_irf.loc[:, 'irf_mbb1'] = IRF_cover68[0, :]
df_irf.loc[:, 'irf_mbb2'] = IRF_cover68[1, :]
df_irf.loc[:, 'irf_mbb_aug1'] = IRF_cover_aug68[0, :]
df_irf.loc[:, 'irf_mbb_aug2'] = IRF_cover_aug68[1, :]


#%% Compute the Coverage Rates of Normalized IRFs

# MBB percentile coverage rates
IRFnorm_cover68 = np.zeros((K, n_imp))
IRFnorm_cover68[:, :] = np.mean(irf_norm_include68, axis=2)

# MBB percentile coverage rates
IRFnorm_cover_aug68 = np.zeros((K, n_imp))
IRFnorm_cover_aug68[:, :] = np.mean(irf_norm_include_aug68, axis=2)

# Anderson Rubin type grid coverage rates
AR_mbb_cover68 = np.zeros((K, n_imp))
AR_mbb_cover68[:, :] = np.mean(ar_mbb_include68, axis=2)

# Lag augmented Anderson Rubin type grid coverage rates
AR_mbb_cover_aug68 = np.zeros((K, n_imp))
AR_mbb_cover_aug68[:, :] = np.mean(ar_mbb_include_aug68, axis=2)

# Store in a pandas dataframe
cols_irfnorm = ['irfnorm_mbb1', 'irfnorm_mbb2',
                'irfnorm_mbb_aug1', 'irfnorm_mbb_aug2',
                'ar_mbb1', 'ar_mbb2', 'ar_mbb_aug1', 'ar_mbb_aug2']
df_irfnorm = pd.DataFrame(index=np.arange(0, n_imp), columns=cols_irfnorm)

df_irfnorm.loc[:, 'irfnorm_mbb1'] = IRFnorm_cover68[0, :]
df_irfnorm.loc[:, 'irfnorm_mbb2'] = IRFnorm_cover68[1, :]
df_irfnorm.loc[:, 'irfnorm_mbb_aug1'] = IRFnorm_cover_aug68[0, :]
df_irfnorm.loc[:, 'irfnorm_mbb_aug2'] = IRFnorm_cover_aug68[1, :]
df_irfnorm.loc[:, 'ar_mbb1'] = AR_mbb_cover68[0, :]
df_irfnorm.loc[:, 'ar_mbb2'] = AR_mbb_cover68[1, :]
df_irfnorm.loc[:, 'ar_mbb_aug1'] = AR_mbb_cover_aug68[0, :]
df_irfnorm.loc[:, 'ar_mbb_aug2'] = AR_mbb_cover_aug68[1, :]


#%% Compute the Coverage Rates of FEVDs

# MBB percentile coverage rates for FEVDs
FEVD_cover68 = np.zeros((K, n_imp))
FEVD_cover68[:, :] = np.mean(fevd_include68, axis=2)

# MBB percentile coverage rates for FEVDs
FEVD_cover_aug68 = np.zeros((K, n_imp))
FEVD_cover_aug68[:, :] = np.mean(fevd_include_aug68, axis=2)

# Store in a pandas dataframe
cols_fevd = ['fevd_mbb1', 'fevd_mbb2', 'fevd_mbb_aug1', 'fevd_mbb_aug2']
df_fevd = pd.DataFrame(index=np.arange(0, n_imp), columns=cols_fevd)

df_fevd.loc[:, 'fevd_mbb1'] = FEVD_cover68[0, :]
df_fevd.loc[:, 'fevd_mbb2'] = FEVD_cover68[1, :]
df_fevd.loc[:, 'fevd_mbb_aug1'] = FEVD_cover_aug68[0, :]
df_fevd.loc[:, 'fevd_mbb_aug2'] = FEVD_cover_aug68[1, :]


#%% Compute the Avg Lengths of Confidence Intervals/Sets for Normalized IRFs

# Computations
len_mbb_avg = np.nanmean(len_mbb, axis=2)
len_mbb_aug_avg = np.nanmean(len_mbb_aug, axis=2)
len_mbb_ar_avg = np.nanmean(len_mbb_ar, axis=2)
len_mbb_ar_aug_avg = np.nanmean(len_mbb_ar_aug, axis=2)
len_mbb_fevd_avg = np.nanmean(len_mbb_fevd, axis=2)
len_mbb_fevd_aug_avg = np.nanmean(len_mbb_fevd_aug, axis=2)

# Store in a pandas dataframe
cols_length = ['irfnorm_mbb1', 'irfnorm_mbb2',
               'irfnorm_mbb_aug1', 'irfnorm_mbb_aug2',
               'ar_mbb1', 'ar_mbb2', 'ar_mbb_aug1', 'ar_mbb_aug2',
               'fevd_mbb1', 'fevd_mbb2', 'fevd_mbb_aug1', 'fevd_mbb_aug2']
df_length = pd.DataFrame(index=np.arange(0, n_imp), columns=cols_length)

df_length.loc[:, 'irfnorm_mbb1'] = len_mbb_avg[0, :]
df_length.loc[:, 'irfnorm_mbb2'] = len_mbb_avg[1, :]
df_length.loc[:, 'irfnorm_mbb_aug1'] = len_mbb_aug_avg[0, :]
df_length.loc[:, 'irfnorm_mbb_aug2'] = len_mbb_aug_avg[1, :]
df_length.loc[:, 'ar_mbb1'] = len_mbb_ar_avg[0, :]
df_length.loc[:, 'ar_mbb2'] = len_mbb_ar_avg[1, :]
df_length.loc[:, 'ar_mbb_aug1'] = len_mbb_ar_aug_avg[0, :]
df_length.loc[:, 'ar_mbb_aug2'] = len_mbb_ar_aug_avg[1, :]
df_length.loc[:, 'fevd_mbb1'] = len_mbb_fevd_avg[0, :]
df_length.loc[:, 'fevd_mbb2'] = len_mbb_fevd_avg[1, :]
df_length.loc[:, 'fevd_mbb_aug1'] = len_mbb_fevd_aug_avg[0, :]
df_length.loc[:, 'fevd_mbb_aug2'] = len_mbb_fevd_aug_avg[1, :]


#%% Save the Results

# Save to Excel
str_file = 'lag_augmented_results.xlsx'
with pd.ExcelWriter(str_file) as writer:
    df_parm.to_excel(writer, sheet_name='parameters')
    df_irf.to_excel(writer, sheet_name='irf')
    df_irfnorm.to_excel(writer, sheet_name='irfnorm')
    df_fevd.to_excel(writer, sheet_name='fevd')
    df_length.to_excel(writer, sheet_name='length')

# Delete temprary variables
del str_file


#%% Disply the Run Time
print(' ')
print('This script finished in ' 
      + str(np.round((time.time()-start)/60, decimals = 1)) + ' minutes.')

