# -*- coding: utf-8 -*-
"""
This script produces Figure G.23 in the appendix of "Asymptotically Valid
Bootstrap Inference for Proxy SVARs" by Carsten Jentsch and Kurt G. Lunsford.

This script saves
    FigureApp23.pdf,
shown as Figure G.23 in the appendix.

This script calls the following Excel file:
    lag_augmented_results.xlsx
This Excel file is produced by appendix_script_augment.py, which must be run
before this script can be used.

This script uses numpy, pandas, and matplotlib packages.
"""

#%% Import Packages and Functions

# Pandas, NumPy, and Matplotlib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


#%% Parameters of the Script

# These parameters must match what is in the Excel files
K = 2           # Number of variables in the VAR
n_imp = 21      # Horizon for IRFs and FEVDs, including the impact (0 horizon)


#%% Load the Data

df_parms = pd.read_excel('lag_augmented_results.xlsx', 'parameters', index_col=0)
df_irfnorm = pd.read_excel('lag_augmented_results.xlsx', 'irfnorm', index_col=0)


#%% Parameters to Print

psi = df_parms.loc[0, 'psi']
n_sample = df_parms.loc[0, 'n_sample']
ell = df_parms.loc[0, 'ell']


#%% Collect the Results

# Array for DGP1
plot_data = np.zeros((K, n_imp, 8))
plot_data[:, :, 0] = df_irfnorm.iloc[:, 0:K].T
plot_data[:, :, 1] = df_irfnorm.iloc[:, 2*K:3*K].T
plot_data[:, :, 2] = np.nan
plot_data[:, :, 3] = np.nan
plot_data[:, :, 4] = np.nan
plot_data[:, :, 5] = np.nan
plot_data[:, :, 6] = df_irfnorm.iloc[:, K:2*K].T
plot_data[:, :, 7] = df_irfnorm.iloc[:, 3*K:4*K].T


#%% Plot the Results

# Variables for the x-axes
x_vars = np.arange(n_imp)
x_tick = np.linspace(0, n_imp-1, int((n_imp-1)/4))

# Markers to be used in the plots
marker_choices = ['x', 'o', 'v', 's', '*', '^', '>', '<']

# Figure and axes instances
fig, ax= plt.subplots(1, 2, figsize=(12, 6))

# First variable
for ii in range(8):
    ax[0].plot(x_vars, plot_data[0, :, ii], linewidth=2, linestyle="-",
               marker=marker_choices[ii], markerfacecolor='None', 
               markeredgewidth=2)
ax[0].plot(x_vars, 0.68*np.ones((n_imp, 1)), color="black",
           linewidth=2, linestyle="-")
ax[0].set_xlim(0, n_imp-1)
ax[0].set_ylim(0, 1)
ax[0].set_xticks(x_tick)
ax[0].set_yticks(np.linspace(0, 1, 6))
ax[0].set_xlabel('Horizon', fontsize=12)
ax[0].set_ylabel('Coverage Rate', fontsize=12)
t = ('$y_{1,t}, T=' + str(n_sample) + ', \ell=' + str(ell)
     + ', \psi=' + str(psi) + '$')
ax[0].set_title(t, fontsize=12)

# Second variable
p0, = ax[1].plot(x_vars, plot_data[1, :, 0], linewidth=2, linestyle="-",
                 marker=marker_choices[0], markerfacecolor='None', 
                 markeredgewidth=2)
p1, = ax[1].plot(x_vars, plot_data[1, :, 1], linewidth=2, linestyle="-",
                 marker=marker_choices[1], markerfacecolor='None', 
                 markeredgewidth=2)
p2, = ax[1].plot(x_vars, plot_data[1, :, 2], linewidth=2, linestyle="-",
                 marker=marker_choices[2], markerfacecolor='None', 
                 markeredgewidth=2)
p3, = ax[1].plot(x_vars, plot_data[1, :, 3], linewidth=2, linestyle="-",
                 marker=marker_choices[3], markerfacecolor='None', 
                 markeredgewidth=2)
p4, = ax[1].plot(x_vars, plot_data[1, :, 4], linewidth=2, linestyle="-",
                 marker=marker_choices[4], markerfacecolor='None', 
                 markeredgewidth=2)
p5, = ax[1].plot(x_vars, plot_data[1, :, 5], linewidth=2, linestyle="-",
                 marker=marker_choices[5], markerfacecolor='None', 
                 markeredgewidth=2)
p6, = ax[1].plot(x_vars, plot_data[1, :, 6], linewidth=2, linestyle="-",
                 marker=marker_choices[6], markerfacecolor='None', 
                 markeredgewidth=2)
p7, = ax[1].plot(x_vars, plot_data[1, :, 7], linewidth=2, linestyle="-",
                 marker=marker_choices[7], markerfacecolor='None', 
                 markeredgewidth=2)
ax[1].plot(x_vars, 0.68*np.ones((n_imp, 1)), color="black",
              linewidth=2, linestyle="-")
ax[1].set_xlim(0, n_imp-1)
ax[1].set_ylim(0, 1)
ax[1].set_xticks(x_tick)
ax[1].set_yticks(np.linspace(0, 1, 6))
ax[1].set_xlabel('Horizon', fontsize=12)
ax[1].set_ylabel('Coverage Rate', fontsize=12)
t = ('$y_{2,t}, T=' + str(n_sample) + ', \ell=' + str(ell) 
     + ', \psi=' + str(psi) + '$')
ax[1].set_title(t, fontsize=12)


ax[1].legend([p0, p1, p6, p7],
             ['MBB', 'grid MBB AR', 'MBB lag augment', 'grid MBB AR lag augment'],
             ncol=4, loc='center', bbox_to_anchor=(-0.1, -0.2), fontsize=12)

plt.savefig('FigureApp23.pdf', bbox_inches='tight')

