# -*- coding: utf-8 -*-
"""
This module contains the functions to estimate a lag-augmented proxy SVAR
model, using one proxy variable to identify one structural shock, and to
compute confidence intervals and sets.

These functions were written for "Asymptotically Valid Bootstrap Inference for 
Proxy SVARs" by Carsten Jentsch and Kurt G. Lunsford.

This module has the following functions:
    estimate_proxy_svar
    make_dynamics
    make_proxy_svar_dynamics_aug
    make_boot_dynamics_aug
    make_ar_vals
    make_ar_set
    make_proxy_svar_mbb_aug

The main functions are
    make_proxy_svar_dynamics : Estimates a lag-augmented proxy SVAR and
                               produces IRFs and FEVDs
    make_proxy_svar_mbb : Uses MBB to produce confidence intervals for
                          lag-augmented proxy SVAR
The other functions exist to support these funtions.

This module uses the numpy and numba packages.
Several functions in this module use the @njit decorator from numba.
"""

import numpy as np
from numba import njit


@njit
def estimate_proxy_svar(yy, xx, mm):
    """
    Uses VAR data and a proxy variable to estimate the parameters of a proxy
    SVAR.
    """
    
    # Dimensions of the data
    TT, KK = yy.shape
    NN = xx[0, :].size
    
    # Estimate VAR coefficients by least squares
    mat1 = np.zeros((NN, NN))
    mat1[:, :] = xx.T @ xx
    
    mat2 = np.zeros((NN, KK))
    mat2[:, :] = xx.T @ yy
    
    A_est = np.zeros((NN, KK))
    A_est = np.linalg.solve(mat1, mat2)
    
    # VAR residuals
    U_est = np.zeros((TT, KK))
    U_est[:, :] = yy - xx @ A_est
    
    # Covariance matrices
    covUU_est = np.zeros((KK, KK))
    covUU_est[:, :] = (U_est.T @ U_est)/TT
    
    covUM_est = np.zeros((KK, 1))
    covUM_est[:, :] = (U_est.T @ mm)/TT
    
    # Proxy SVAR identification
    phi_sq_est = np.zeros((1, 1))
    phi_sq_est[:, :] = covUM_est.T @ np.linalg.solve(covUU_est, covUM_est)
    phi_est = np.sqrt(phi_sq_est[0, 0])
    H1_est = covUM_est/phi_est
    
    # Transpose A
    A_est = A_est.T
    
    return A_est, U_est, covUU_est, covUM_est, H1_est


@njit
def make_dynamics(AA, CovUU, CovUM, HH1, pp, normscale, num_imp):
    """
    Uses the estimated parameters of a proxy SVAR to produce IRFs and FEVDs.
    """
    
    # Number of VAR variables
    KK = AA[:, 0].size
    
    # Procude the companion matrix
    comp = np.zeros((KK*pp, KK*pp))
    if pp == 1:
        comp[:, :] = AA[:, 1:KK+1]
    elif pp > 1:
        comp[0:KK, :] = AA[:, 1:KK*pp+1]
        comp[KK:KK*pp, 0:KK*(pp-1)] = np.eye(KK*(pp-1))
    
    # Arrays to store IRFs, FEVDs, and statistics for AR type confidence set
    irf = np.empty((KK, num_imp))
    fevd = np.empty((KK, num_imp))
    svma = np.empty((KK, num_imp))
    
    # IRF at horizon 0
    irf[:, 0] = HH1[:, 0]
    
    # FEVD at horizon 0
    numer = np.empty((KK, 1))
    denom = CovUU
    for kk in range(KK):
        numer[kk, 0] = HH1[kk, 0]*HH1[kk, 0]
        fevd[kk, 0] = numer[kk, 0]/denom[kk, kk]
    
    # Statistic for Anderson Rubin type confidence set
    svma[:, 0] = CovUM[:, 0]
    
    # Top KK rows of companion matrix raised to 0
    comp_exp = np.zeros((KK, KK*pp))
    comp_exp[0:KK, 0:KK] = np.eye(KK)
    
    # Loop to produce IRFs and FEVDs at horizons greater than 1
    for ii in range(num_imp-1):
        # Top KK rows of companion matrix raised to ii+1
        comp_exp = comp_exp @ comp
        
        # Store the VMA coefficients
        Phi = np.empty((KK, KK))
        Phi[:, :] = comp_exp[0:KK, 0:KK]
        
        # Compute the one standard deviation IRF
        temp = np.zeros((KK, 1))
        for nn in range(KK):
            for kk in range(KK):
                temp[nn, 0] = temp[nn, 0] + Phi[nn, kk]*HH1[kk, 0]
        irf[:, ii+1] = temp[:, 0]
        
        # Compute the FEVD
        denom = denom + (Phi @ CovUU) @ Phi.T
        for kk in range(KK):
            numer[kk, 0] = numer[kk, 0] + temp[kk, 0]*temp[kk, 0]
            fevd[kk, ii+1] = numer[kk, 0]/denom[kk, kk]
        
        # Construct structural VMA for the structural shock of interest
        temp = np.zeros((KK, 1))
        for nn in range(KK):
            for kk in range(KK):
                temp[nn, 0] = temp[nn, 0] + Phi[nn, kk]*CovUM[kk, 0]
        svma[:, ii+1] = temp[:, 0]
    
    #Compute normalized IRF
    irf_norm = normscale*irf/irf[0, 0]
    
    return irf, irf_norm, fevd, svma


def make_proxy_svar_dynamics_aug(y_in, m_in, pp, ss, hh):
    """
    Estimates a loag-augmented proxy SVAR model, using one proxy variable to
    identify one structural shock, and produces F statistics, IRFs, and FEVDs.
    
    This function calls:
        estimate_proxy_svar
        make_dynamics
    """
    
    # Dimensions of the data
    n_samp, n_var = y_in.shape
    
    # Number of augmented lags
    pp_aug = pp + 1
    
    # Effective number of observations
    n_eff = n_samp - pp_aug
    
    # Adjust the sample of proxy variables for the number of VAR lags
    m_use = np.zeros((n_eff, 1))
    m_use[:, 0] = m_in[pp_aug:n_samp, 0]
    
    # Construct the matrix of left-hand side variables
    lhs_aug = np.empty((n_eff, n_var))
    lhs_aug[:, :] = y_in[pp_aug:n_samp, :]
    
    # Construct the matrix of right-hand side variables
    rhs_aug = np.ones((n_eff, n_var*pp_aug+1))
    for ii in range(pp_aug):
        rhs_aug[:, n_var*ii+1:n_var*(ii+1)+1] = y_in[pp_aug-ii-1:n_samp-ii-1, :]
    
    # Estimate the proxy SVAR
    A_hat, U_hat, SigmaU_hat, SigmaUM_hat, H1_hat \
        = estimate_proxy_svar(lhs_aug, rhs_aug, m_use)
    
    # Collect the output
    dict_out = {'a_hat' : A_hat,
                'u_hat' : U_hat,
                'sigma_uu_hat' : SigmaU_hat,
                'sigma_um_hat' : SigmaUM_hat,
                'h1_hat' : H1_hat}
    
    # Compute the F statistic
    rhs = np.ones((n_eff, (1+n_var)))
    rhs[:, 1:1+n_var] = U_hat
    
    Pi = np.zeros((1+n_var, 1))
    Pi[:, :] = np.linalg.inv(rhs.T @ rhs) @ (rhs.T @ m_use)
    
    err = np.zeros((n_eff, 1))
    err[:, :] = m_use - rhs @ Pi
    
    err_cls = np.zeros((n_eff, 1))
    err_cls[:, :] = m_use - np.mean(m_use[:, 0])
    
    sse_cls = np.zeros((1, 1))
    sse_cls[:, :] = err_cls.T @ err_cls
    
    sse = np.zeros((1, 1))
    sse[:, :] = err.T @ err
    
    F = (n_eff - 1)*((sse_cls[0, 0]/sse[0, 0]) - 1)/n_var
    dict_out.update({'f_stat' : F})
   
    # Compute the F_iv statistic
    rhs = np.ones((n_eff, 2))
    rhs[:, 1:2] = m_use
    
    Pi = np.zeros((2, 1))
    Pi[:, :] = np.linalg.inv(rhs.T @ rhs) @ (rhs.T @ U_hat[:, 0:1])
    
    err = np.zeros((n_eff, 1))
    err[:, :] = U_hat[:, 0:1] - rhs @ Pi
    
    err_cls = np.zeros((n_eff, 1))
    err_cls[:, :] = U_hat[:, 0:1]
    
    sse_cls = np.zeros((1, 1))
    sse_cls[:, :] = err_cls.T @ err_cls
    
    sse = np.zeros((1, 1))
    sse[:, :] = err.T @ err
    
    F_iv = (n_eff - 1)*((sse_cls[0, 0]/sse[0, 0]) - 1)
    dict_out.update({'f_iv_stat' : F_iv})
    
    # Compute the IRFs, FEVDs, and statistics for AR-type confidence sets
    A_alt = np.zeros((n_var, n_var*pp+1))
    A_alt[:, :] = A_hat[:, 0:n_var*pp+1]
    IRF_hat, IRFnorm_hat, FEVD_hat, SVMA_hat \
        = make_dynamics(A_alt, SigmaU_hat, SigmaUM_hat, H1_hat, pp, ss, hh)
    
    # Collect the output
    dict_out.update({'irf_hat' : IRF_hat,
                     'irf_norm_hat' : IRFnorm_hat,
                     'fevd_hat' : FEVD_hat,
                     'svma_hat' : SVMA_hat})
    
    return dict_out


@njit
def make_boot_dynamics_aug(AA, u_in, m_in, y_init, nlags, snorm, numimp):
    """
    Uses resampled VAR residuals and a resampled proxy variable to compute
    bootstrapped IRFs and FEVDs, based on a lag-augmented VAR.
    
    This function calls:
        estimate_proxy_svar
        make_dynamics
    """
    
    # Dimensions of the data
    TT, KK = u_in.shape
    
    # Loop to construct the VAR variables
    x_star = np.zeros((TT, KK*(nlags+1)+1))
    x_star[0, :] = y_init[:]
    y_star = u_in
    for ii in range(TT-1):
        # Compute hist @ AA
        for nn in range(KK):
            for kk in range(KK*(nlags+1)+1):
                y_star[ii, nn] = y_star[ii, nn] + x_star[ii, kk]*AA[kk, nn]
        
        # Update X_star
        x_star[ii+1, KK+1:KK*(nlags+1)+1] = x_star[ii, 1:KK*nlags+1]
        x_star[ii+1, 1:KK+1] = y_star[ii, :]
        x_star[ii+1, 0] = 1
    
    # Last observation of VAR variables
    for nn in range(KK):
        for kk in range(KK*(nlags+1)+1):
            y_star[TT-1, nn] = y_star[TT-1, nn] + x_star[TT-1, kk]*AA[kk, nn]
    
    # Estimate the proxy SVAR
    a_star, u_starhat, covUU_star, covUM_star, h1_star \
        = estimate_proxy_svar(y_star, x_star, m_in)
    
    # Compute the IRFs, FEVDs, and statistics for AR-type confidence sets
    a_small = np.zeros((KK, KK*nlags+1))
    a_small[:, :] = a_star[:, 0:KK*nlags+1]
    irf_star, irf_norm_star, fevd_star, svma_star \
        = make_dynamics(a_small, covUU_star, covUM_star, h1_star,
                        nlags, snorm, numimp)
    
    return irf_star, irf_norm_star, fevd_star, svma_star


@njit
def make_ar_vals(vals1, vals2):
    """
    Computes useful values for the grid AR confidence sets, computed with
    either the MBB or with simulation.
    """
    
    # Dimensions of the arrays
    k_dim, irf_dim = vals1.shape
    grid_dim = vals2.size
    
    # Output array
    vals_out = np.zeros((k_dim, irf_dim, grid_dim))
    
    # Loops for computations
    for kk in range(k_dim):
        for nn in range(irf_dim):
            for ggg in range(grid_dim):
                vals_out[kk, nn, ggg] = vals1[kk, nn] - vals2[ggg]
    
    return vals_out


@njit
def make_ar_set(rates, grid, scale):
    """
    This function determines which grid points are included in the Anderson
    Rubin type confidence set.
    """
    
    # Dimensions of the input variable
    numVAR, numImp, numGrid = rates.shape
    
    # Storage arrays
    index68 = np.empty((numGrid, numImp, numVAR))
    index90 = np.empty((numGrid, numImp, numVAR))
    index95 = np.empty((numGrid, numImp, numVAR))
    
    # Loop to construct the confidence sets
    for kk in range(numVAR):
        for ii in range(numImp):
            for ggg in range(numGrid):
                # Check if grid point is included in confidence set
                check = np.empty(3)
                check[0] = grid[ggg] == scale
                check[1] = ii == 0
                check[2] = kk == 0
                if np.all(check) == 1:
                    index68[ggg, ii, kk] = True
                    index95[ggg, ii, kk] = True
                else:
                    rr = rates[kk, ii, ggg]
                    
                    temp68 = np.empty(numVAR)
                    temp68[0] = 0.16 <= rr
                    temp68[1] = rr <= 0.84
                    index68[ggg, ii, kk] = np.all(temp68)
                    
                    temp90 = np.empty(numVAR)
                    temp90[0] = 0.05 <= rr
                    temp90[1] = rr <= 0.95
                    index90[ggg, ii, kk] = np.all(temp90)
                    
                    temp95 = np.empty(numVAR)
                    temp95[0] = 0.025 <= rr
                    temp95[1] = rr <= 0.975
                    index95[ggg, ii, kk] = np.all(temp95)
    
    return index68, index90, index95


def make_proxy_svar_mbb_aug(AA, UU, MM, Yinit, nlags, snorm, numimp, numboot,
                            length, toggleAR=0, ar_grid=None, svma_vals=None):
    """
    Runs the moving block bootstrap and computes confidence intervals for a
    lag-augmented VAR.
    
    This function calls:
        make_dynamics
        make_boot_dynamics_aug
        make_ar_vals
        make_ar_set
    """
    
    # Dimensions of the data
    TT, KK = UU.shape
    
    # Transpose AA for convenience
    AA = AA.T
    
    # Construct the blocks
    numblocks = TT - length + 1
    u_blocks = np.zeros((length, KK, numblocks))
    m_blocks = np.zeros((length, 1, numblocks))
    for ii in range(numblocks):
        u_blocks[:, :, ii] = UU[ii:length+ii, :]
        m_blocks[:, :, ii] = MM[ii:length+ii, :]
    
    # Centering for the blocks
    u_temp = np.zeros((length, KK))
    m_temp = np.zeros((length, 1))
    for ii in range(length):
        m_temp[ii, 0] = np.mean(MM[ii:TT-length+ii+1, 0])
        for nn in range(KK):
            u_temp[ii, nn] = np.mean(UU[ii:TT-length+ii+1, nn])
    
    numResample = int(np.ceil(TT/length))
    u_center = np.zeros((numResample*length, KK))
    for ii in range(numResample):
        u_center[ii*length:(ii+1)*length, :] = u_temp
    
    # Set dimensions of arrays to hold bootstrapped estimates
    irf_store = np.zeros((KK, numimp, numboot))
    irf_norm_store = np.zeros((KK, numimp, numboot))
    fevd_store = np.zeros((KK, numimp, numboot))
    
    # Set up the grid and storage array for AR type confidence sets if needed
    if toggleAR == 1:
        # Check if an Anderson Rubin grid is provided
        if ar_grid is None:
            print(' ')
            print('No Anderson Rubin grid is provided.')
            print('The grid is set to "numpy.linspace(-10, 10, num=201)".')
            print(' ')
            ar_grid = np.zeros(201)
            ar_grid[:] = np.linspace(-10, 10, num=201)
        else:
            # Flatten the AR grid
            ar_grid = ar_grid.flatten()
        
        # Storage array
        numgrid = ar_grid.size
        ar_store = np.zeros((KK, numimp, numgrid, numboot))
        
        # Check if structural vector moving average values are provided
        if svma_vals is None:
            # Covariance matrices
            covUU_hat = np.zeros((KK, KK))
            covUU_hat[:, :] = (UU.T @ UU)/TT
            
            covUM_hat = np.zeros((KK, 1))
            covUM_hat[:, :] = (UU.T @ MM)/TT
            
            # Proxy SVAR identification
            phi_sq_hat = np.zeros((1, 1))
            phi_sq_hat[:, :] = covUM_hat.T @ np.linalg.solve(covUU_hat, covUM_hat)
            phi_hat = np.sqrt(phi_sq_hat[0, 0])
            H1_hat = covUM_hat/phi_hat
            
            AA_small = np.zeros((len(AA[:, 0])-KK, KK))
            irf_vals, irf_norm_vals, fevd_vals, svma_vals \
                = make_dynamics(AA_small.T, covUU_hat, covUM_hat, H1_hat,
                                nlags, snorm, numimp)
    
    # Bootstrap loop
    for boot in range(numboot):
        # Resample the blocks
        ratio = int(np.ceil(TT/length))
        u_temp = np.empty((length*ratio, KK))
        m_temp = np.empty((length*ratio, 1))
        for ii in range(ratio):
            index = np.random.randint(numblocks)
            u_temp[ii*length:(ii+1)*length, :] = u_blocks[:, :, index]
            m_temp[ii*length:(ii+1)*length, :] = m_blocks[:, :, index]
        
        # Center the re-sampled variables
        u_temp[:, :] = u_temp - u_center
        
        # Truncate to get bootstrapped U and M
        u_star = np.empty((TT, KK))
        u_star[:, :] = u_temp[0:TT, :]
        m_star = np.empty((TT, 1))
        m_star[:, :] = m_temp[0:TT, :]
        
        # Compute bootstrapped IRFs and FEVDs
        irf_boot, irf_norm_boot, fevd_boot, svma_boot = \
            make_boot_dynamics_aug(AA, u_star, m_star, Yinit.flatten(),
                                   nlags, snorm, numimp)
        
        # Store the bootstrapped IRFs and FEVDs
        irf_store[:, :, boot] = irf_boot
        irf_norm_store[:, :, boot] = irf_norm_boot
        fevd_store[:, :, boot] = fevd_boot
        
        # Compute and store statistics for Anderson Rubin type confidence set
        if toggleAR == 1:
            temp1 = np.zeros((KK, numimp))
            temp1[:, :] = snorm*svma_boot
            
            temp2 = np.zeros(numgrid)
            temp2[:] = svma_boot[0, 0]*ar_grid
            
            ar_store[:, :, :, boot] = make_ar_vals(temp1, temp2)
    
    # Sort the output
    irf_sort = np.sort(irf_store, axis=2)
    irf_norm_sort = np.sort(irf_norm_store, axis=2)
    fevd_sort = np.sort(fevd_store, axis=2)
    
    # Compute the 68 percent percentile confidence intervals
    num16 = int(round(0.16*numboot))
    num84 = int(round(0.84*numboot))
    ci68_irf = np.empty((2, numimp, KK))
    ci68_irf_norm = np.empty((2, numimp, KK))
    ci68_fevd = np.empty((2, numimp, KK))
    for kk in range(KK):
        ci68_irf[0, :, kk] = irf_sort[kk, :, num16]
        ci68_irf[1, :, kk] = irf_sort[kk, :, num84]
        
        ci68_irf_norm[0, :, kk] = irf_norm_sort[kk, :, num16]
        ci68_irf_norm[1, :, kk] = irf_norm_sort[kk, :, num84]
        
        ci68_fevd[0, :, kk] = fevd_sort[kk, :, num16]
        ci68_fevd[1, :, kk] = fevd_sort[kk, :, num84]
    
    # Compute the 90 percent percentile confidence intervals
    num05 = int(round(0.05*numboot))
    num95 = int(round(0.95*numboot))
    ci90_irf = np.empty((2, numimp, KK))
    ci90_irf_norm = np.empty((2, numimp, KK))
    ci90_fevd = np.empty((2, numimp, KK))
    for kk in range(KK):
        ci90_irf[0, :, kk] = irf_sort[kk, :, num05]
        ci90_irf[1, :, kk] = irf_sort[kk, :, num95]
        
        ci90_irf_norm[0, :, kk] = irf_norm_sort[kk, :, num05]
        ci90_irf_norm[1, :, kk] = irf_norm_sort[kk, :, num95]
        
        ci90_fevd[0, :, kk] = fevd_sort[kk, :, num05]
        ci90_fevd[1, :, kk] = fevd_sort[kk, :, num95]
    
    # Compute the 95 percent percentile confidence intervals
    num025 = int(round(0.025*numboot))
    num975 = int(round(0.975*numboot))
    ci95_irf = np.empty((2, numimp, KK))
    ci95_irf_norm = np.empty((2, numimp, KK))
    ci95_fevd = np.empty((2, numimp, KK))
    for kk in range(KK):
        ci95_irf[0, :, kk] = irf_sort[kk, :, num025]
        ci95_irf[1, :, kk] = irf_sort[kk, :, num975]
        
        ci95_irf_norm[0, :, kk] = irf_norm_sort[kk, :, num025]
        ci95_irf_norm[1, :, kk] = irf_norm_sort[kk, :, num975]
        
        ci95_fevd[0, :, kk] = fevd_sort[kk, :, num025]
        ci95_fevd[1, :, kk] = fevd_sort[kk, :, num975]
    
    # Store the output in a dictionary
    mbb_dict = {'ci68_irf' : ci68_irf,
                'ci90_irf' : ci90_irf,
                'ci95_irf' : ci95_irf,
                'ci68_irf_norm' : ci68_irf_norm,
                'ci90_irf_norm' : ci90_irf_norm,
                'ci95_irf_norm' : ci95_irf_norm,
                'ci68_fevd' : ci68_fevd,
                'ci90_fevd' : ci90_fevd,
                'ci95_fevd' : ci95_fevd}
    
    # Compute AR confidence sets if needed
    if toggleAR == 1:
        rates = np.zeros((KK, numimp, numgrid))
        rates[:, :, :] = np.mean((ar_store <= 0), axis=3)
        
        ar_index68, ar_index90, ar_index95 = make_ar_set(rates, ar_grid, snorm)
        mbb_dict.update({'ar_index68' : ar_index68,
                         'ar_index90' : ar_index90,
                         'ar_index95' : ar_index95,
                         'ar_grid': ar_grid})
    
    return mbb_dict
