# -*- coding: utf-8 -*-
"""
This module contains two functions used for the Monte Carlo simulations in
"Asymptotically Valid Bootstrap Inference for Proxy SVARs" by Carsten Jentsch 
and Kurt G. Lunsford.

It has the following functions:
    make_data_sv
    check_include

This module uses the numpy and numba packages.
One function in this module uses the @njit decorator from numba.
"""

# Load the packages
import numpy as np
from numba import njit


def make_data_sv(TT, KK, pp, AA, HH, parm, mu, std_e):
    """
    Produces simulated VAR data with innovations that have stochastic
    volatility as well as noise for a proxy variable.
    """
    
    # Innovations to volatility
    ee = np.zeros((TT, KK))
    ee[:, :] = std_e*np.random.standard_normal(size=(TT, KK))
    
    # Adjust mu for the number of variables in the VAR
    mu_big = mu*np.ones(KK)
        
    # Simulate the log of sigma
    log_sig = np.zeros((TT, KK))
    log_sig[0, :] = (1 - parm)*mu_big + ee[0, :]
    for ii in range(TT-1):
        log_sig[ii+1, :] = (1 - parm)*mu_big + parm*log_sig[ii, :] + ee[ii+1, :]
    
    # The structural shocks
    ww = np.random.standard_normal(size=(TT, KK))
    eps = np.zeros((TT, KK))
    eps[:, :] = np.multiply(ww, np.exp(log_sig))
    
    # The VAR innovations
    uu = np.zeros((TT, KK))
    uu[:, :] = eps @ HH.T
    
    # Storage arrays for variables of interest
    YY = np.empty((TT, KK))                 # observable variables in the VAR
    
    # Construct the observable variables in the VAR
    history = np.zeros((1, KK*pp+1))        # set history of variables to zero
    history[0, 0] = 1                       # set the constant
    for ii in range(TT):
        # Compute the observable variables
        YY[ii, :] = history @ AA.T + uu[ii, :]
        
        # Update the history of VAR variables
        history[0, KK+1:KK*pp+1] = history[0, 1:KK*(pp-1)+1]
        history[0, 1:KK+1] = YY[ii, :]
    
    # Additional noise for constructing the proxy variables
    noise = np.random.standard_normal(size=(TT, 1))
    
    return YY, eps, noise


@njit
def check_include(KK, numImp, CI, values):
    """
    Checks if the true IRFs or FEVDs are included in the confidence intervals
    or sets.
    """
    include = np.zeros((KK, numImp))
    temp = np.empty((KK, 1))
    for nn in range(KK):
        for hh in range(numImp):
            temp[0, 0] = CI[0, hh, nn] <= values[nn, hh]
            temp[1, 0] = values[nn, hh] <= CI[1, hh, nn]
            include[nn, hh] = np.all(temp)
    
    return include

