# -*- coding: utf-8 -*-
"""
This script produces Figure 3 in Section 5 of "Asymptotically Valid Bootstrap
Inference for Proxy SVARs" by Carsten Jentsch and Kurt G. Lunsford.

This script saves
    Figure3.pdf

This script calls the following Excel files:
    dgp1_results.xlsx
    dgp2_results.xlsx
    dgp3_results.xlsx
These Excel files are produced by main_script.py, which must be run three times
(one time to produce each of the three Excel files) before this script can be
used.

This script uses numpy, pandas, and matplotlib packages.
"""

#%% Import Packages and Functions

# Pandas, NumPy, and Matplotlib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


#%% Parameters of the Script

# These parameters must match what is in the Excel files
K = 2           # Number of variables in the VAR
n_imp = 21      # Horizon for IRFs and FEVDs, including the impact (0 horizon)


#%% Load the Data

df_parms1 = pd.read_excel('dgp1_results.xlsx', 'parameters', index_col=0)
df_fevd1 = pd.read_excel('dgp1_results.xlsx', 'fevd', index_col=0)

df_parms2 = pd.read_excel('dgp2_results.xlsx', 'parameters', index_col=0)
df_fevd2 = pd.read_excel('dgp2_results.xlsx', 'fevd', index_col=0)

df_parms3 = pd.read_excel('dgp3_results.xlsx', 'parameters', index_col=0)
df_fevd3 = pd.read_excel('dgp3_results.xlsx', 'fevd', index_col=0)


#%% Parameters to Print

psi1 = df_parms1.loc[0, 'psi']
psi2 = df_parms2.loc[0, 'psi']
psi3 = df_parms3.loc[0, 'psi']

n_sample = df_parms1.loc[0, 'n_sample']
ell = df_parms1.loc[0, 'ell']


#%% Collect the Results

# Array for DGP1
plot_data1 = np.zeros((K, n_imp))
plot_data1[:, :] = df_fevd1.iloc[:, 0:K].T

# Array for DGP2
plot_data2 = np.zeros((K, n_imp))
plot_data2[:, :] = df_fevd2.iloc[:, 0:K].T

# Array for DGP3
plot_data3 = np.zeros((K, n_imp))
plot_data3[:, :] = df_fevd3.iloc[:, 0:K].T


#%% Plot the Results

# Variables for the x-axes
x_vars = np.arange(n_imp)
x_tick = np.linspace(0, n_imp-1, int((n_imp-1)/4))

# Markers to be used in the plots
marker_choices = ['x', 'o', 'v', 's']

# Figure and axes instances
fig, ax= plt.subplots(2, 3, figsize=(14, 11))

# First variable with dgp1
ax[0, 0].plot(x_vars, plot_data1[0, :], linewidth=2, linestyle="-",
              marker=marker_choices[0], markerfacecolor='None', 
              markeredgewidth=2)
ax[0, 0].plot(x_vars, 0.68*np.ones((n_imp, 1)), color="black",
              linewidth=2, linestyle="-")
ax[0, 0].set_xlim(0, n_imp-1)
ax[0, 0].set_ylim(0, 1)
ax[0, 0].set_xticks(x_tick)
ax[0, 0].set_yticks(np.linspace(0, 1, 6))
ax[0, 0].set_xlabel('Horizon', fontsize=12)
ax[0, 0].set_ylabel('Coverage Rate', fontsize=12)
t = ('$y_{1,t}, T=' + str(n_sample) + ', \ell=' + str(ell) 
     + ', \psi=' + str(psi1) + '$')
ax[0, 0].set_title(t, fontsize=12)

# Second variable with dgp1
ax[1, 0].plot(x_vars, plot_data1[1, :], linewidth=2, linestyle="-",
              marker=marker_choices[0], markerfacecolor='None', 
              markeredgewidth=2)
ax[1, 0].plot(x_vars, 0.68*np.ones((n_imp, 1)), color="black",
              linewidth=2, linestyle="-")
ax[1, 0].set_xlim(0, n_imp-1)
ax[1, 0].set_ylim(0, 1)
ax[1, 0].set_xticks(x_tick)
ax[1, 0].set_yticks(np.linspace(0, 1, 6))
ax[1, 0].set_xlabel('Horizon', fontsize=12)
ax[1, 0].set_ylabel('Coverage Rate', fontsize=12)
t = ('$y_{2,t}, T=' + str(n_sample) + ', \ell=' + str(ell) 
     + ', \psi=' + str(psi1) + '$')
ax[1, 0].set_title(t, fontsize=12)

# First variable with dgp2
ax[0, 1].plot(x_vars, plot_data2[0, :], linewidth=2, linestyle="-",
              marker=marker_choices[0], markerfacecolor='None', 
              markeredgewidth=2)
ax[0, 1].plot(x_vars, 0.68*np.ones((n_imp, 1)), color="black",
              linewidth=2, linestyle="-")
ax[0, 1].set_xlim(0, n_imp-1)
ax[0, 1].set_ylim(0, 1)
ax[0, 1].set_xticks(x_tick)
ax[0, 1].set_yticks(np.linspace(0, 1, 6))
ax[0, 1].set_xlabel('Horizon', fontsize=12)
ax[0, 1].set_ylabel('Coverage Rate', fontsize=12)
t = ('$y_{1,t}, T=' + str(n_sample) + ', \ell=' + str(ell) 
     + ', \psi=' + str(psi2) + '$')
ax[0, 1].set_title(t, fontsize=12)

# Second variable with dgp2
ax[1, 1].plot(x_vars, plot_data2[1, :], linewidth=2, linestyle="-",
              marker=marker_choices[0], markerfacecolor='None', 
              markeredgewidth=2)
ax[1, 1].plot(x_vars, 0.68*np.ones((n_imp, 1)), color="black",
              linewidth=2, linestyle="-")
ax[1, 1].set_xlim(0, n_imp-1)
ax[1, 1].set_ylim(0, 1)
ax[1, 1].set_xticks(x_tick)
ax[1, 1].set_yticks(np.linspace(0, 1, 6))
ax[1, 1].set_xlabel('Horizon', fontsize=12)
ax[1, 1].set_ylabel('Coverage Rate', fontsize=12)
t = ('$y_{2,t}, T=' + str(n_sample) + ', \ell=' + str(ell) 
     + ', \psi=' + str(psi2) + '$')
ax[1, 1].set_title(t, fontsize=12)

# First variable with dgp3
ax[0, 2].plot(x_vars, plot_data3[0, :], linewidth=2, linestyle="-",
              marker=marker_choices[0], markerfacecolor='None', 
              markeredgewidth=2)
ax[0, 2].plot(x_vars, 0.68*np.ones((n_imp, 1)), color="black",
              linewidth=2, linestyle="-")
ax[0, 2].set_xlim(0, n_imp-1)
ax[0, 2].set_ylim(0, 1)
ax[0, 2].set_xticks(x_tick)
ax[0, 2].set_yticks(np.linspace(0, 1, 6))
ax[0, 2].set_xlabel('Horizon', fontsize=12)
ax[0, 2].set_ylabel('Coverage Rate', fontsize=12)
t = ('$y_{1,t}, T=' + str(n_sample) + ', \ell=' + str(ell) 
     + ', \psi=' + str(psi3) + '$')
ax[0, 2].set_title(t, fontsize=12)

# Second variable with dgp3
ax[1, 2].plot(x_vars, plot_data3[1, :], linewidth=2, linestyle="-",
              marker=marker_choices[0], markerfacecolor='None', 
              markeredgewidth=2)
ax[1, 2].plot(x_vars, 0.68*np.ones((n_imp, 1)), color="black",
              linewidth=2, linestyle="-")
ax[1, 2].set_xlim(0, n_imp-1)
ax[1, 2].set_ylim(0, 1)
ax[1, 2].set_xticks(x_tick)
ax[1, 2].set_yticks(np.linspace(0, 1, 6))
ax[1, 2].set_xlabel('Horizon', fontsize=12)
ax[1, 2].set_ylabel('Coverage Rate', fontsize=12)
t = ('$y_{2,t}, T=' + str(n_sample) + ', \ell=' + str(ell) 
     + ', \psi=' + str(psi3) + '$')
ax[1, 2].set_title(t, fontsize=12)

plt.savefig('Figure3.pdf', bbox_inches='tight')

