# -*- coding: utf-8 -*-
"""
This module contains the functions to estimate a proxy SVAR model, using one 
proxy variable to identify one structural shock, and to compute confidence
intervals and sets.

These functions were written for "Asymptotically Valid Bootstrap Inference for 
Proxy SVARs" by Carsten Jentsch and Kurt G. Lunsford.

This module has the following functions:
    estimate_proxy_svar
    make_dynamics
    make_proxy_svar_dynamics
    make_boot_dynamics
    make_ar_vals
    make_ar_set
    make_proxy_svar_mbb
    make_proxy_svar_wbr
    make_halls
    make_msw_cs
    make_ar_sim

The main functions are
    make_proxy_svar_dynamics : Estimates a proxy SVAR and produces IRFs and FEVDs
    make_proxy_svar_mbb : Uses MBB to produce confidence intervals
    make_proxy_svar_wbr : Uses a Rademacher wild bootstrap to produce confidence intervals (not recommended)
    make_halls : Turns percentile intervals into Hall's percentile intervals 
    make_msw_cs : Produces Montiel Olea, Stock and Watson confidence sets
    make_ar_sim : Produces grid Anderson-Rubin confidence sets with simulation
The other functions exist to support these funtions.

This module uses the numpy and numba packages.
Several functions in this module use the @njit decorator from numba.
"""

# Load the packages
import numpy as np
from numba import njit


@njit
def estimate_proxy_svar(yy, xx, mm):
    """
    Uses VAR data and a proxy variable to estimate the parameters of a proxy
    SVAR.
    """
    
    # Dimensions of the data
    TT, KK = yy.shape
    NN = xx[0, :].size
    
    # Estimate VAR coefficients by least squares
    mat1 = np.zeros((NN, NN))
    mat1[:, :] = xx.T @ xx
    
    mat2 = np.zeros((NN, KK))
    mat2[:, :] = xx.T @ yy
    
    A_est = np.zeros((NN, KK))
    A_est = np.linalg.solve(mat1, mat2)
    
    # VAR residuals
    U_est = np.zeros((TT, KK))
    U_est[:, :] = yy - xx @ A_est
    
    # Covariance matrices
    covUU_est = np.zeros((KK, KK))
    covUU_est[:, :] = (U_est.T @ U_est)/TT
    
    covUM_est = np.zeros((KK, 1))
    covUM_est[:, :] = (U_est.T @ mm)/TT
    
    # Proxy SVAR identification
    phi_sq_est = np.zeros((1, 1))
    phi_sq_est[:, :] = covUM_est.T @ np.linalg.solve(covUU_est, covUM_est)
    phi_est = np.sqrt(phi_sq_est[0, 0])
    H1_est = covUM_est/phi_est
    
    # Transpose A
    A_est = A_est.T
    
    return A_est, U_est, covUU_est, covUM_est, H1_est


@njit
def make_dynamics(AA, CovUU, CovUM, HH1, pp, normscale, num_imp):
    """
    Uses the estimated parameters of a proxy SVAR to produce IRFs and FEVDs.
    """
    
    # Number of VAR variables
    KK = AA[:, 0].size
    
    # Procude the companion matrix
    comp = np.zeros((KK*pp, KK*pp))
    if pp == 1:
        comp[:, :] = AA[:, 1:KK+1]
    elif pp > 1:
        comp[0:KK, :] = AA[:, 1:KK*pp+1]
        comp[KK:KK*pp, 0:KK*(pp-1)] = np.eye(KK*(pp-1))
    
    # Arrays to store IRFs, FEVDs, and statistics for AR type confidence set
    irf = np.empty((KK, num_imp))
    fevd = np.empty((KK, num_imp))
    svma = np.empty((KK, num_imp))
    
    # IRF at horizon 0
    irf[:, 0] = HH1[:, 0]
    
    # FEVD at horizon 0
    numer = np.empty((KK, 1))
    denom = CovUU
    for kk in range(KK):
        numer[kk, 0] = HH1[kk, 0]*HH1[kk, 0]
        fevd[kk, 0] = numer[kk, 0]/denom[kk, kk]
    
    # Statistic for Anderson Rubin type confidence set
    svma[:, 0] = CovUM[:, 0]
    
    # Top KK rows of companion matrix raised to 0
    comp_exp = np.zeros((KK, KK*pp))
    comp_exp[0:KK, 0:KK] = np.eye(KK)
    
    # Loop to produce IRFs and FEVDs at horizons greater than 1
    for ii in range(num_imp-1):
        # Top KK rows of companion matrix raised to ii+1
        comp_exp = comp_exp @ comp
        
        # Store the VMA coefficients
        Phi = np.empty((KK, KK))
        Phi[:, :] = comp_exp[0:KK, 0:KK]
        
        # Compute the one standard deviation IRF
        temp = np.zeros((KK, 1))
        for nn in range(KK):
            for kk in range(KK):
                temp[nn, 0] = temp[nn, 0] + Phi[nn, kk]*HH1[kk, 0]
        irf[:, ii+1] = temp[:, 0]
        
        # Compute the FEVD
        denom = denom + (Phi @ CovUU) @ Phi.T
        for kk in range(KK):
            numer[kk, 0] = numer[kk, 0] + temp[kk, 0]*temp[kk, 0]
            fevd[kk, ii+1] = numer[kk, 0]/denom[kk, kk]
        
        # Construct structural VMA for the structural shock of interest
        temp = np.zeros((KK, 1))
        for nn in range(KK):
            for kk in range(KK):
                temp[nn, 0] = temp[nn, 0] + Phi[nn, kk]*CovUM[kk, 0]
        svma[:, ii+1] = temp[:, 0]
    
    #Compute normalized IRF
    irf_norm = normscale*irf/irf[0, 0]
    
    return irf, irf_norm, fevd, svma


def make_proxy_svar_dynamics(y_in, m_in, pp, ss, hh):
    """
    Estimates a proxy SVAR model, using one proxy variable to identify one
    structural shock, and produces F statistics, IRFs, and FEVDs.
    
    This function calls:
        estimate_proxy_svar
        make_dynamics
    """
    
    # Dimensions of the data
    n_samp, n_var = y_in.shape
    
    # Effective number of observations
    n_eff = n_samp - pp
    
    # Adjust the sample of proxy variables for the number of VAR lags
    m_use = np.zeros((n_eff, 1))
    m_use[:, 0] = m_in[pp:n_samp, 0]
    
    # Construct the matrix of left-hand side variables
    lhs = np.empty((n_eff, n_var))
    lhs[:, :] = y_in[pp:n_samp, :]
    
    # Construct the matrix of right-hand side variables
    rhs = np.ones((n_eff, n_var*pp+1))
    for ii in range(pp):
        rhs[:, n_var*ii+1:n_var*(ii+1)+1] = y_in[pp-ii-1:n_samp-ii-1, :]
    
    # Estimate the proxy SVAR
    A_hat, U_hat, SigmaU_hat, SigmaUM_hat, H1_hat \
        = estimate_proxy_svar(lhs, rhs, m_use)
    
    # Collect the output
    dict_out = {'a_hat' : A_hat,
                'u_hat' : U_hat,
                'sigma_uu_hat' : SigmaU_hat,
                'sigma_um_hat' : SigmaUM_hat,
                'h1_hat' : H1_hat}
    
    # Compute the F statistic
    rhs = np.ones((n_eff, (1+n_var)))
    rhs[:, 1:1+n_var] = U_hat
    
    Pi = np.zeros((1+n_var, 1))
    Pi[:, :] = np.linalg.inv(rhs.T @ rhs) @ (rhs.T @ m_use)
    
    err = np.zeros((n_eff, 1))
    err[:, :] = m_use - rhs @ Pi
    
    err_cls = np.zeros((n_eff, 1))
    err_cls[:, :] = m_use - np.mean(m_use[:, 0])
    
    sse_cls = np.zeros((1, 1))
    sse_cls[:, :] = err_cls.T @ err_cls
    
    sse = np.zeros((1, 1))
    sse[:, :] = err.T @ err
    
    F = (n_eff - 1)*((sse_cls[0, 0]/sse[0, 0]) - 1)/n_var
    dict_out.update({'f_stat' : F})
   
    # Compute the F_iv statistic
    rhs = np.ones((n_eff, 2))
    rhs[:, 1:2] = m_use
    
    Pi = np.zeros((2, 1))
    Pi[:, :] = np.linalg.inv(rhs.T @ rhs) @ (rhs.T @ U_hat[:, 0:1])
    
    err = np.zeros((n_eff, 1))
    err[:, :] = U_hat[:, 0:1] - rhs @ Pi
    
    err_cls = np.zeros((n_eff, 1))
    err_cls[:, :] = U_hat[:, 0:1]
    
    sse_cls = np.zeros((1, 1))
    sse_cls[:, :] = err_cls.T @ err_cls
    
    sse = np.zeros((1, 1))
    sse[:, :] = err.T @ err
    
    F_iv = (n_eff - 1)*((sse_cls[0, 0]/sse[0, 0]) - 1)
    dict_out.update({'f_iv_stat' : F_iv})
    
    # Compute the IRFs, FEVDs, and statistics for AR-type confidence sets
    IRF_hat, IRFnorm_hat, FEVD_hat, SVMA_hat \
        = make_dynamics(A_hat, SigmaU_hat, SigmaUM_hat, H1_hat, pp, ss, hh)
    
    # Collect the output
    dict_out.update({'irf_hat' : IRF_hat,
                     'irf_norm_hat' : IRFnorm_hat,
                     'fevd_hat' : FEVD_hat,
                     'svma_hat' : SVMA_hat})
    
    return dict_out


@njit
def make_boot_dynamics(AA, u_in, m_in, y_init, nlags, snorm, numimp):
    """
    Uses resampled VAR residuals and a resampled proxy variable to compute
    bootstrapped IRFs and FEVDs.
    
    This function calls:
        estimate_proxy_svar
        make_dynamics
    """
    
    # Dimensions of the data
    TT, KK = u_in.shape

    # Loop to construct the VAR variables
    x_star = np.zeros((TT, KK*nlags+1))
    x_star[0, :] = y_init[:]
    y_star = u_in
    for ii in range(TT-1):
        # Compute hist @ AA
        for nn in range(KK):
            for kk in range(KK*nlags+1):
                y_star[ii, nn] = y_star[ii, nn] + x_star[ii, kk]*AA[kk, nn]
        
        # Update X_star
        x_star[ii+1, KK+1:KK*nlags+1] = x_star[ii, 1:KK*(nlags-1)+1]
        x_star[ii+1, 1:KK+1] = y_star[ii, :]
        x_star[ii+1, 0] = 1
    
    # Last observation of VAR variables
    for nn in range(KK):
        for kk in range(KK*nlags+1):
            y_star[TT-1, nn] = y_star[TT-1, nn] + x_star[TT-1, kk]*AA[kk, nn]
    
    # Estimate the proxy SVAR
    a_star, u_starhat, covUU_star, covUM_star, h1_star \
        = estimate_proxy_svar(y_star, x_star, m_in)
    
    # Compute the IRFs, FEVDs, and statistics for AR-type confidence sets
    irf_star, irf_norm_star, fevd_star, svma_star \
        = make_dynamics(a_star, covUU_star, covUM_star, h1_star, nlags, snorm, numimp)
    
    return irf_star, irf_norm_star, fevd_star, svma_star


@njit
def make_ar_vals(vals1, vals2):
    """
    Computes useful values for the grid AR confidence sets, computed with
    either the MBB or with simulation.
    """
    
    # Dimensions of the arrays
    k_dim, irf_dim = vals1.shape
    grid_dim = vals2.size
    
    # Output array
    vals_out = np.zeros((k_dim, irf_dim, grid_dim))
    
    # Loops for computations
    for kk in range(k_dim):
        for nn in range(irf_dim):
            for ggg in range(grid_dim):
                vals_out[kk, nn, ggg] = vals1[kk, nn] - vals2[ggg]
    
    return vals_out


@njit
def make_ar_set(rates, grid, scale):
    """
    This function determines which grid points are included in the Anderson
    Rubin type confidence set.
    """
    
    # Dimensions of the input variable
    numVAR, numImp, numGrid = rates.shape
    
    # Storage arrays
    index68 = np.empty((numGrid, numImp, numVAR))
    index90 = np.empty((numGrid, numImp, numVAR))
    index95 = np.empty((numGrid, numImp, numVAR))
    
    # Loop to construct the confidence sets
    for kk in range(numVAR):
        for ii in range(numImp):
            for ggg in range(numGrid):
                # Check if grid point is included in confidence set
                check = np.empty(3)
                check[0] = grid[ggg] == scale
                check[1] = ii == 0
                check[2] = kk == 0
                if np.all(check) == 1:
                    index68[ggg, ii, kk] = True
                    index95[ggg, ii, kk] = True
                else:
                    rr = rates[kk, ii, ggg]
                    
                    temp68 = np.empty(numVAR)
                    temp68[0] = 0.16 <= rr
                    temp68[1] = rr <= 0.84
                    index68[ggg, ii, kk] = np.all(temp68)
                    
                    temp90 = np.empty(numVAR)
                    temp90[0] = 0.05 <= rr
                    temp90[1] = rr <= 0.95
                    index90[ggg, ii, kk] = np.all(temp90)
                    
                    temp95 = np.empty(numVAR)
                    temp95[0] = 0.025 <= rr
                    temp95[1] = rr <= 0.975
                    index95[ggg, ii, kk] = np.all(temp95)
    
    return index68, index90, index95


def make_proxy_svar_mbb(AA, UU, MM, Yinit, nlags, snorm, numimp, numboot,
                        length, toggleAR=0, ar_grid=None, svma_vals=None):
    """
    Runs the moving block bootstrap and computes confidence intervals.
    
    This function calls:
        make_dynamics
        make_boot_dynamics
        make_ar_vals
        make_ar_set
    """
    
    # Dimensions of the data
    TT, KK = UU.shape
    
    # Transpose AA for convenience
    AA = AA.T
    
    # Construct the blocks
    numblocks = TT - length + 1
    u_blocks = np.zeros((length, KK, numblocks))
    m_blocks = np.zeros((length, 1, numblocks))
    for ii in range(numblocks):
        u_blocks[:, :, ii] = UU[ii:length+ii, :]
        m_blocks[:, :, ii] = MM[ii:length+ii, :]
    
    # Centering for the blocks
    u_temp = np.zeros((length, KK))
    m_temp = np.zeros((length, 1))
    for ii in range(length):
        m_temp[ii, 0] = np.mean(MM[ii:TT-length+ii+1, 0])
        for nn in range(KK):
            u_temp[ii, nn] = np.mean(UU[ii:TT-length+ii+1, nn])
    
    numResample = int(np.ceil(TT/length))
    u_center = np.zeros((numResample*length, KK))
    for ii in range(numResample):
        u_center[ii*length:(ii+1)*length, :] = u_temp
    
    # Set dimensions of arrays to hold bootstrapped estimates
    irf_store = np.zeros((KK, numimp, numboot))
    irf_norm_store = np.zeros((KK, numimp, numboot))
    fevd_store = np.zeros((KK, numimp, numboot))
    
    # Set up the grid and storage array for AR type confidence sets if needed
    if toggleAR == 1:
        # Check if an Anderson Rubin grid is provided
        if ar_grid is None:
            print(' ')
            print('No Anderson Rubin grid is provided.')
            print('The grid is set to "numpy.linspace(-10, 10, num=201)".')
            print(' ')
            ar_grid = np.zeros(201)
            ar_grid[:] = np.linspace(-10, 10, num=201)
        else:
            # Flatten the AR grid
            ar_grid = ar_grid.flatten()
        
        # Storage array
        numgrid = ar_grid.size
        ar_store = np.zeros((KK, numimp, numgrid, numboot))
        
        # Check if structural vector moving average values are provided
        if svma_vals is None:
            # Covariance matrices
            covUU_hat = np.zeros((KK, KK))
            covUU_hat[:, :] = (UU.T @ UU)/TT
            
            covUM_hat = np.zeros((KK, 1))
            covUM_hat[:, :] = (UU.T @ MM)/TT
            
            # Proxy SVAR identification
            phi_sq_hat = np.zeros((1, 1))
            phi_sq_hat[:, :] = covUM_hat.T @ np.linalg.solve(covUU_hat, covUM_hat)
            phi_hat = np.sqrt(phi_sq_hat[0, 0])
            H1_hat = covUM_hat/phi_hat
            
            irf_vals, irf_norm_vals, fevd_vals, svma_vals \
                = make_dynamics(AA.T, covUU_hat, covUM_hat, H1_hat, nlags, snorm, numimp)
    
    # Bootstrap loop
    for boot in range(numboot):
        # Resample the blocks
        ratio = int(np.ceil(TT/length))
        u_temp = np.empty((length*ratio, KK))
        m_temp = np.empty((length*ratio, 1))
        for ii in range(ratio):
            index = np.random.randint(numblocks)
            u_temp[ii*length:(ii+1)*length, :] = u_blocks[:, :, index]
            m_temp[ii*length:(ii+1)*length, :] = m_blocks[:, :, index]
        
        # Center the re-sampled variables
        u_temp[:, :] = u_temp - u_center
        
        # Truncate to get bootstrapped U and M
        u_star = np.empty((TT, KK))
        u_star[:, :] = u_temp[0:TT, :]
        m_star = np.empty((TT, 1))
        m_star[:, :] = m_temp[0:TT, :]
        
        # Compute bootstrapped IRFs and FEVDs
        irf_boot, irf_norm_boot, fevd_boot, svma_boot = \
            make_boot_dynamics(AA, u_star, m_star, Yinit.flatten(),
                               nlags, snorm, numimp)
        
        # Store the bootstrapped IRFs and FEVDs
        irf_store[:, :, boot] = irf_boot
        irf_norm_store[:, :, boot] = irf_norm_boot
        fevd_store[:, :, boot] = fevd_boot
        
        # Compute and store statistics for Anderson Rubin type confidence set
        if toggleAR == 1:
            temp1 = np.zeros((KK, numimp))
            temp1[:, :] = snorm*svma_boot
            
            temp2 = np.zeros(numgrid)
            temp2[:] = svma_boot[0, 0]*ar_grid
            
            ar_store[:, :, :, boot] = make_ar_vals(temp1, temp2)
    
    # Sort the output
    irf_sort = np.sort(irf_store, axis=2)
    irf_norm_sort = np.sort(irf_norm_store, axis=2)
    fevd_sort = np.sort(fevd_store, axis=2)
    
    # Compute the 68 percent percentile confidence intervals
    num16 = int(round(0.16*numboot))
    num84 = int(round(0.84*numboot))
    ci68_irf = np.empty((2, numimp, KK))
    ci68_irf_norm = np.empty((2, numimp, KK))
    ci68_fevd = np.empty((2, numimp, KK))
    for kk in range(KK):
        ci68_irf[0, :, kk] = irf_sort[kk, :, num16]
        ci68_irf[1, :, kk] = irf_sort[kk, :, num84]
        
        ci68_irf_norm[0, :, kk] = irf_norm_sort[kk, :, num16]
        ci68_irf_norm[1, :, kk] = irf_norm_sort[kk, :, num84]
        
        ci68_fevd[0, :, kk] = fevd_sort[kk, :, num16]
        ci68_fevd[1, :, kk] = fevd_sort[kk, :, num84]
    
    # Compute the 95 percent percentile confidence intervals
    num025 = int(round(0.025*numboot))
    num975 = int(round(0.975*numboot))
    ci95_irf = np.empty((2, numimp, KK))
    ci95_irf_norm = np.empty((2, numimp, KK))
    ci95_fevd = np.empty((2, numimp, KK))
    for kk in range(KK):
        ci95_irf[0, :, kk] = irf_sort[kk, :, num025]
        ci95_irf[1, :, kk] = irf_sort[kk, :, num975]
        
        ci95_irf_norm[0, :, kk] = irf_norm_sort[kk, :, num025]
        ci95_irf_norm[1, :, kk] = irf_norm_sort[kk, :, num975]
        
        ci95_fevd[0, :, kk] = fevd_sort[kk, :, num025]
        ci95_fevd[1, :, kk] = fevd_sort[kk, :, num975]
    
    # Store the output in a dictionary
    mbb_dict = {'ci68_irf' : ci68_irf,
                'ci95_irf' : ci95_irf,
                'ci68_irf_norm' : ci68_irf_norm,
                'ci95_irf_norm' : ci95_irf_norm,
                'ci68_fevd' : ci68_fevd,
                'ci95_fevd' : ci95_fevd}
    
    # Compute AR confidence sets if needed
    if toggleAR == 1:
        rates = np.zeros((KK, numimp, numgrid))
        rates[:, :, :] = np.mean((ar_store <= 0), axis=3)
        
        ar_index68, ar_index90, ar_index95 = make_ar_set(rates, ar_grid, snorm)
        mbb_dict.update({'ar_index68' : ar_index68,
                         'ar_index90' : ar_index90,
                         'ar_index95' : ar_index95,
                         'ar_grid': ar_grid})
    
    return mbb_dict


def make_proxy_svar_wbr(AA, UU, MM, Yinit, nlags, snorm, numimp, numboot):
    """
    Runs the Rademacher wild bootstrap and computes confidence intervals.
    !! Not recommended !!
    
    This functions calls:
        make_boot_dynamics
    """
    
    # Dimensions of the data
    TT, KK = UU.shape
    
    # Transpose AA for convenience
    AA = AA.T
    
    # Set dimensions of arrays to hold bootstrapped estimates
    irf_store = np.zeros((KK, numimp, numboot))
    irf_norm_store = np.zeros((KK, numimp, numboot))
    fevd_store = np.zeros((KK, numimp, numboot))
    
    # Bootstrap loop
    for boot in range(numboot):
        # Generate the multipliers
        multiplier = 2*np.random.randint(2, size=TT) - 1
        
        # Resample the residuals
        u_star = np.zeros((TT, KK))
        for kk in range(KK):
            u_star[:, kk] = np.multiply(multiplier, UU[:, kk])
        
        m_star = np.zeros((TT, 1))
        m_star[:, 0] = np.multiply(multiplier, MM[:, 0])
        
        # Compute bootstrapped IRFs and FEVDs
        irf_boot, irf_norm_boot, fevd_boot, svma_boot = \
            make_boot_dynamics(AA, u_star, m_star, Yinit.flatten(),
                               nlags, snorm, numimp)
        
        # Store the bootstrapped IRFs and FEVDs
        irf_store[:, :, boot] = irf_boot
        irf_norm_store[:, :, boot] = irf_norm_boot
        fevd_store[:, :, boot] = fevd_boot
    
    # Sort the output
    irf_sort = np.sort(irf_store, axis=2)
    irf_norm_sort = np.sort(irf_norm_store, axis=2)
    fevd_sort = np.sort(fevd_store, axis=2)
    
    # Compute the 68 percent percentile confidence intervals
    num16 = int(round(0.16*numboot))
    num84 = int(round(0.84*numboot))
    ci68_irf = np.empty((2, numimp, KK))
    ci68_irf_norm = np.empty((2, numimp, KK))
    ci68_fevd = np.empty((2, numimp, KK))
    for kk in range(KK):
        ci68_irf[0, :, kk] = irf_sort[kk, :, num16]
        ci68_irf[1, :, kk] = irf_sort[kk, :, num84]
        
        ci68_irf_norm[0, :, kk] = irf_norm_sort[kk, :, num16]
        ci68_irf_norm[1, :, kk] = irf_norm_sort[kk, :, num84]
        
        ci68_fevd[0, :, kk] = fevd_sort[kk, :, num16]
        ci68_fevd[1, :, kk] = fevd_sort[kk, :, num84]
    
    # Compute the 95 percent percentile confidence intervals
    num025 = int(round(0.025*numboot))
    num975 = int(round(0.975*numboot))
    ci95_irf = np.empty((2, numimp, KK))
    ci95_irf_norm = np.empty((2, numimp, KK))
    ci95_fevd = np.empty((2, numimp, KK))
    for kk in range(KK):
        ci95_irf[0, :, kk] = irf_sort[kk, :, num025]
        ci95_irf[1, :, kk] = irf_sort[kk, :, num975]
        
        ci95_irf_norm[0, :, kk] = irf_norm_sort[kk, :, num025]
        ci95_irf_norm[1, :, kk] = irf_norm_sort[kk, :, num975]
        
        ci95_fevd[0, :, kk] = fevd_sort[kk, :, num025]
        ci95_fevd[1, :, kk] = fevd_sort[kk, :, num975]
    
    # Store the output in a dictionary
    wbr_dict = {'ci68_irf' : ci68_irf,
                'ci95_irf' : ci95_irf,
                'ci68_irf_norm' : ci68_irf_norm,
                'ci95_irf_norm' : ci95_irf_norm,
                'ci68_fevd' : ci68_fevd,
                'ci95_fevd' : ci95_fevd}
    
    return wbr_dict


def make_halls(pctile_interval, irf):
    """
    Converts percentile intervals into Hall's percentile intervals.
    """
    
    # Dimensions of the data
    ignore, numimp, KK = pctile_interval.shape
    
    # Hall's percentile interval
    halls_interval = np.empty((2, numimp, KK))
    for kk in range(KK):
        halls_interval[0, :, kk] = 2*irf[kk, :] - pctile_interval[1, :, kk]
        halls_interval[1, :, kk] = 2*irf[kk, :] - pctile_interval[0, :, kk]
    
    return halls_interval


def make_msw_cs(AA, YY, UU, MM, nlags, snorm, numimp):
    """
    Produces the analytic confidence sets in Montiel Olea, Stock and Watson.
    These confidence sets only apply to normalized IRFs.
    """
    
    # Dimensions of the VAR data
    TT, KK = UU.shape
    
    # Number of variables to simulate
    dim = KK*(KK*nlags + 1) + KK
    
    # Construct the right-hand side matrix
    rhs = np.ones((TT, 1+KK*nlags))
    for nn in range(nlags):
        rhs[:, 1+nlags*nn:1+nlags*(nn+1)] = YY[nlags-nn-1:TT+nlags-nn-1, :]
    
    # Products of the right-hand side variables and the VAR residuals
    zz = np.empty((TT, dim))
    counter = 0
    for ii in range(KK*nlags+1):
        for kk in range(KK):
            zz[:, counter] = np.multiply(UU[:, kk], rhs[:, ii])
            counter = counter + 1
    
    # Products of proxy variables and the VAR residuals less the mean
    gamma_est = (MM.T @ UU)/TT
    for kk in range(KK):
        zz[:, KK*(KK*nlags+1)+kk] = (np.multiply(UU[:, kk], MM[:, 0]) 
                                     - gamma_est[0, kk])
    
    # I am using heteroskedasticity robust covariance here
    # Newey and West (1987) or other HAR covariance can also be used
    omega = np.zeros((dim, dim))
    omega = zz.T @ zz/TT
    
    # Auxiliary matrix
    I_K = np.eye(KK)
    q1 = rhs.T @ rhs/TT
    q1_inv = np.linalg.inv(q1)
    q2 = (MM.T @ rhs/TT) @ q1_inv
    aux = np.eye(KK*(KK*nlags+1)+KK)
    aux[0:KK*(KK*nlags+1), 0:KK*(KK*nlags+1)] = np.kron(q1_inv, I_K)
    aux[KK*(KK*nlags+1):dim, 0:KK*(KK*nlags+1)] = -np.kron(q2, I_K)
    
    # Variance to be used for MSW analytic confidence set
    v = np.zeros((dim, dim))
    v[:, :] = (aux @ omega) @ aux.T
    v11 = v[KK:KK*(KK*nlags+1), KK:KK*(KK*nlags+1)]
    v31 = v[KK*(KK*nlags+1):dim, KK:KK*(KK*nlags+1)]
    v33 = v[KK*(KK*nlags+1):dim, KK*(KK*nlags+1):dim]
    
    # Produce the companion matrix
    comp = np.zeros((KK*nlags, KK*nlags))
    if nlags == 1:
        comp[:, :] = AA[:, 1:KK+1]
    elif nlags > 1:
        comp[0:KK, :] = AA[:, 1:KK*nlags+1]
        comp[KK:KK*nlags, 0:KK*(nlags-1)] = np.eye(KK*(nlags-1))
    
    # Storage arrays
    bigA = np.empty((KK*nlags, KK*nlags, numimp))
    G = np.zeros((KK*KK, KK*KK*nlags, numimp+1))
    
    # Values for horizon = 1
    bigA[:, :, 0] = np.eye(KK*nlags)
    G[:, :, 1] = np.kron(bigA[0:KK, :, 0], I_K)
    
    # Companion matrix raised to 0
    comp_exp = np.eye((KK*nlags))
    
    # Compute the G matrices
    for ii in range(numimp-1):
        # Companion matrix raised to ii-1
        comp_exp[:, :] = (comp.T) @ comp_exp
        
        # Update the bigA matrix
        bigA[:, :, ii+1] = comp_exp
        
        # Compute the G matrix
        temp = np.zeros((KK*KK, KK*KK*nlags))
        for nn in range(ii+2):
            Phi = bigA[0:KK, 0:KK, nn].T
            temp = temp + np.kron(bigA[0:KK, :, ii+1-nn], Phi)
        G[:, :, ii+2] = temp
    
    # Wald statistic
    wald = TT*gamma_est[0, 0]*gamma_est[0, 0]/v33[0, 0]
    
    # Chi squared critical values
    crit68 = 0.9889     # alpha = 0.68
    crit95 = 3.8415     # alpha = 0.95
    
    # Set up the 68 percent confidence set
    aa68 = TT*gamma_est[0, 0]*gamma_est[0, 0] - crit68*v33[0, 0]
    if aa68 > 0:
        # If aa68 > 0, then confidence set is bounded
        con_set68 = np.zeros((2, numimp, KK))
    else:
        # If aa68 < 0, then confidence set is unbounded and possibly disjoint
        con_set68 = np.zeros((4, numimp, KK))
    
    # Set up the 95 percent confidence set    
    aa95 = TT*gamma_est[0, 0]*gamma_est[0, 0] - crit95*v33[0, 0]
    if aa95 > 0:
        # If aa95 > 0, then confidence set is bounded
        con_set95 = np.zeros((2, numimp, KK))
    else:
        # If aa95 < 0, then confidence set is unbounded and possibly disjoint
        con_set95 = np.zeros((4, numimp, KK))
    
    # Loop to fill in the confidence sets
    for nn in range(KK):
        # Compute a useful matrices
        Mkron = np.kron(gamma_est, I_K[nn:nn+1, :])
        for ii in range(numimp):
            # Some useful matrices
            Phi = bigA[0:KK, 0:KK, ii].T
            temp1 = Phi[nn:nn+1, :] @ gamma_est.T
            temp2 = Mkron @ G[:, :, ii]
            
            # Compute the value of b
            bpart1 = -TT*temp1*gamma_est[0, 0]
            bpart2 = temp2 @ v31[0:1, :].T + Phi[nn:nn+1, :] @ v33[:, 0:1]
            bb68 = 2*snorm*(bpart1 + crit68*bpart2)
            bb95 = 2*snorm*(bpart1 + crit95*bpart2)
            
            # Compute the value of c
            cpart1 = TT*temp1*temp1
            cpart2 = ((temp2 @ v11) @ temp2.T
                      + 2*(temp2 @ v31.T) @ Phi[nn:nn+1, :].T
                      + (Phi[nn:nn+1, :] @ v33) @ Phi[nn:nn+1, :].T)
            cc68 = snorm*snorm*(cpart1 - crit68*cpart2)
            cc95 = snorm*snorm*(cpart1 - crit95*cpart2)
            
            # Compute the 68 percent confidence set
            temp = bb68*bb68 - 4*aa68*cc68
            if aa68 > 0:
                if temp > 0:
                    con_set68[0, ii, nn] = (-bb68 - np.sqrt(temp))/(2*aa68)
                    con_set68[1, ii, nn] = (-bb68 + np.sqrt(temp))/(2*aa68)
            else:
                con_set68[0, ii, nn] = -np.inf
                con_set68[3, ii, nn] = np.inf
                if temp > 0:
                    con_set68[1, ii, nn] = (-bb68 + np.sqrt(temp))/(2*aa68)
                    con_set68[2, ii, nn] = (-bb68 - np.sqrt(temp))/(2*aa68)
            
            # Compute the 95 percent confidence set
            temp = bb95*bb95 - 4*aa95*cc95
            if aa95 > 0:
                if temp > 0:
                    con_set95[0, ii, nn] = (-bb95 - np.sqrt(temp))/(2*aa95)
                    con_set95[1, ii, nn] = (-bb95 + np.sqrt(temp))/(2*aa95)
            else:
                con_set95[0, ii, nn] = -np.inf
                con_set95[3, ii, nn] = np.inf
                if temp > 0:
                    con_set95[1, ii, nn] = (-bb95 + np.sqrt(temp))/(2*aa95)
                    con_set95[2, ii, nn] = (-bb95 - np.sqrt(temp))/(2*aa95)
    
    # 68% confidence set for normalized impact
    if aa68 > 0:
        con_set68[0, 0, 0] = snorm
        con_set68[1, 0, 0] = snorm
    else:
        con_set68[0, 0, 0] = snorm
        con_set68[1, 0, 0] = snorm
        con_set68[2, 0, 0] = 0
        con_set68[3, 0, 0] = 0
    
    # 95% confidence set for normalized impact
    if aa95 > 0:
        con_set95[0, 0, 0] = snorm
        con_set95[1, 0, 0] = snorm
    else:
        con_set95[0, 0, 0] = snorm
        con_set95[1, 0, 0] = snorm
        con_set95[2, 0, 0] = 0
        con_set95[3, 0, 0] = 0
    
    # Store the output in a dictionary
    msw_dict = {'cs68_irf_norm' : con_set68,
                'cs95_irf_norm' : con_set95,
                'vcv_matrix' : v,
                'wald_stat' : wald}
    
    return msw_dict


def make_ar_sim(YY, MM, nlags, snorm, numimp, numboot, ar_grid):
    """
    Produces Anderson and Rubinc confidene sets based on simulating the
    reduced-form parameters from the estimated covariance matrix.
    """
    
    # Dimensions of the VAR data
    TT, KK = YY.shape
    TT = TT - nlags
    
    # Number of variables to simulate
    dim = KK*(KK*nlags + 1) + KK
        
    # Construct the left-hand side matrix
    lhs = np.ones((TT, KK))
    lhs[:, :] = YY[nlags:TT+nlags, :]
    
    # Construct the right-hand side matrix
    rhs = np.ones((TT, 1+KK*nlags))
    for nn in range(nlags):
        rhs[:, 1+nlags*nn:1+nlags*(nn+1)] = YY[nlags-nn-1:TT+nlags-nn-1, :]
    
    # Least squares regression
    AA = np.zeros((KK, 1+KK*nlags))
    AA[:, :] = (lhs.T @ rhs) @ np.linalg.inv(rhs.T @ rhs)
    
    # VAR residuals
    UU = np.zeros((TT, KK))
    UU[:, :] = lhs - rhs @ AA.T
    
    # Products of the right-hand side variables and the VAR residuals
    zz = np.empty((TT, dim))
    counter = 0
    for ii in range(KK*nlags+1):
        for kk in range(KK):
            zz[:, counter] = np.multiply(UU[:, kk], rhs[:, ii])
            counter = counter + 1
    
    # Products of proxy variables and the VAR residuals less the mean
    gamma_est = (MM[nlags:TT+nlags, :].T @ UU)/TT
    for kk in range(KK):
        zz[:, KK*(KK*nlags+1)+kk] \
            = np.multiply(UU[:, kk], MM[nlags:TT+nlags, 0]) - gamma_est[0, kk]
    
    # I am using heteroskedasticity robust covariance here
    # Newey and West (1987) or other HAR covariance can also be used
    omega = np.zeros((dim, dim))
    omega[:, :] = zz.T @ zz/TT
    
    # Auxiliary matrix
    I_K = np.eye(KK)
    q1 = rhs.T @ rhs/TT
    q1_inv = np.linalg.inv(q1)
    q2 = (MM[nlags:TT+nlags, :].T @ rhs/TT) @ q1_inv
    aux = np.eye(KK*(KK*nlags+1)+KK)
    aux[0:KK*(KK*nlags+1), 0:KK*(KK*nlags+1)] = np.kron(q1_inv, I_K)
    aux[KK*(KK*nlags+1):dim, 0:KK*(KK*nlags+1)] = -np.kron(q2, I_K)
    
    # Variance to draw variables from
    v = np.zeros((dim, dim))
    v[:, :] = ((aux @ omega) @ aux.T)/TT
    
    # Cholesky decomposition of variance
    v_lower = np.linalg.cholesky(v)
    
    # Populate the estimates
    estimates = np.zeros((dim, 1))
    estimates[0:KK*(KK*nlags+1), 0] = AA.T.flatten()
    estimates[KK*(KK*nlags+1):dim, 0] = gamma_est[0, :]
    
    # Simulated draws
    draws = np.zeros((dim, numboot))
    draws[:, :] = (estimates @ np.ones((1, numboot))
                   + v_lower @ np.random.standard_normal(size=(dim, numboot)))
    
    # Proxy SVAR identification -- this is a placeholder and not ultimately used
    sigma_est = np.zeros((KK, KK))
    sigma_est[:, :] = (UU.T @ UU)/TT
    
    phi_sq_est = np.zeros((1, 1))
    phi_sq_est[:, :] = gamma_est @ np.linalg.solve(sigma_est, gamma_est.T)
    
    phi_est = np.sqrt(phi_sq_est[0, 0])
    h1_est = gamma_est.T/phi_est
    
    # Loop to compute vales for Anderson and Rubin confidence set
    ar_grid = ar_grid.flatten()
    numgrid = len(ar_grid)
    ar_store = np.zeros((KK, numimp, numgrid, numboot))
    for boot in range(numboot):
        # The VAR slope coefficients
        a_sim = np.zeros((KK, 1+KK*nlags))
        for jj in range(1+KK*nlags):
            a_sim[:, jj] = draws[KK*jj:KK*(jj+1), boot]
        
        # The covariance of VAR innovations and the proxy
        gamma_sim = np.zeros((KK, 1))
        gamma_sim[:, 0] = draws[KK*(KK*nlags+1):dim, boot]
        
        irf_sim, irf_norm_sim, fevd_sim, svma_sim = \
            make_dynamics(a_sim, sigma_est, gamma_sim, h1_est,
                          nlags, snorm, numimp)
        
        # Compute and store statistics for Anderson Rubin type confidence set
        temp1 = np.zeros((KK, numimp))
        temp1[:, :] = snorm*svma_sim
        
        temp2 = np.zeros(numgrid)
        temp2[:] = svma_sim[0, 0]*ar_grid
        
        ar_store[:, :, :, boot] = make_ar_vals(temp1, temp2)
    
    # Compute the confidence set
    rates = np.zeros((KK, numimp, numgrid))
    rates[:, :, :] = np.mean((ar_store <= 0), axis=3)
    
    ar_index68, ar_index90, ar_index95 = make_ar_set(rates, ar_grid, snorm)
    arsim_dict = {'ar_index68' : ar_index68,
                  'ar_index90' : ar_index90,
                  'ar_index95' : ar_index95,
                  'ar_grid': ar_grid}
    
    return arsim_dict

